from dataclasses import dataclass, field
import os
import re
from tempfile import TemporaryDirectory
import traceback
from typing import Dict, List, Optional, Tuple, Union

import click
from click.testing import CliRunner
import pytest

from anyscale._private.docgen.__main__ import ALL_MODULES, generate
from anyscale._private.docgen.generator import MarkdownGenerator, Module
from anyscale._private.docgen.generator_legacy import parse_legacy_sdks
from anyscale._private.models import ModelBase, ModelEnum
from anyscale._private.sdk import sdk_command
from anyscale.commands.util import AnyscaleCommand, LegacyAnyscaleCommand


def test_full_smoke_test():
    runner = CliRunner()
    with TemporaryDirectory() as d:
        result = runner.invoke(generate, [d])
        # should have an output per module
        expected_output_count = len(ALL_MODULES)

        if result.exception:
            e = result.exception
            stacktrace = traceback.format_exception(type(e), e, e.__traceback__)
            formatted_exception = "".join(stacktrace)

        assert result.exit_code == 0, result.output or formatted_exception
        assert result.output.count("Writing output file") == expected_output_count


def test_parse_legacy_sdks():
    _TEST_API_MD = """\
some random text here

### fake_legacy_sdk

this is a fake legacy [sdk](./models.md#fakesdk)
"""

    _TEST_MODEL_MD = """\
this is random

## FakeLegacyModel

this is a fake legacy [model](#fakemodel)

# next section
"""

    with TemporaryDirectory() as tmp:
        with open(os.path.join(tmp, "api.md"), "w") as f:
            f.write(_TEST_API_MD)

        with open(os.path.join(tmp, "model.md"), "w") as f:
            f.write(_TEST_MODEL_MD)

        legacy_sdks, legacy_models = parse_legacy_sdks(
            os.path.join(tmp, "api.md"), os.path.join(tmp, "model.md"),
        )
        assert legacy_sdks[0].name == "fake_legacy_sdk"
        assert (
            legacy_sdks[0].docstring == "this is a fake legacy [sdk](#fakesdk-legacy)"
        )
        assert legacy_models[0].name == "FakeLegacyModel"
        assert (
            legacy_models[0].docstring
            == "this is a fake legacy [model](#fakemodel-legacy)"
        )


def _fake_module(
    *,
    name: str = "Fake",
    models: Optional[List] = None,
    cli_commands: Optional[List] = None,
    sdk_commands: Optional[List] = None,
    cli_command_group_prefix: Optional[Dict[click.Command, str]] = None,
    legacy_cli_commands: Optional[List] = None,
) -> Module:
    return Module(
        title=name,
        filename=f"{name.lower()}.md",
        cli_prefix=f"anyscale {name.lower()}",
        cli_commands=cli_commands or [],
        sdk_prefix=f"anyscale.{name.lower()}",
        sdk_commands=sdk_commands or [],
        models=models or [],
        cli_command_group_prefix=cli_command_group_prefix,
        legacy_cli_prefix=f"anyscale legacy {name.lower()}",
        legacy_cli_commands=legacy_cli_commands or [],
    )


class TestGenerateModelType:
    def test_basic(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            foo: str = field(metadata={"docstring": "Foo docstring."})

            # Private fields should not be documented.
            _private: str

            __doc_yaml_example__ = """\
config: file"""

            __doc_py_example__ = """\
run_python_command()
"""

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeConfig])]
        ).generate()
        assert len(output) == 1
        assert "## `FakeConfig`" in output["fake.md"]
        assert "FakeConfig does really special things." in output["fake.md"]
        assert "**`foo` (str)**: Foo docstring." in output["fake.md"]
        assert "_private" not in output["fake.md"]

        # Check that model methods are added.
        assert "def __init__(self, **fields) -> FakeConfig" in output["fake.md"]
        assert "def options(self, **fields) -> FakeConfig" in output["fake.md"]
        assert "def to_dict(self) -> Dict[str, Any]" in output["fake.md"]

        # Check that examples were generated.
        assert '<TabItem value="yamlconfig" label="YAML">' in output["fake.md"]
        assert "config: file" in output["fake.md"]
        assert '<TabItem value="pythonsdk" label="Python">' in output["fake.md"]
        assert "run_python_command()" in output["fake.md"]

    def test_missing_field_docstring(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            foo: str

        with pytest.raises(
            ValueError,
            match="Model 'FakeConfig' is missing a docstring for field 'foo'",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_missing_py_example(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_yaml_example__ = "fake: config"

        with pytest.raises(
            ValueError, match="Model 'FakeConfig' is missing a '__doc_py_example__'",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_invalid_py_example(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_yaml_example__ = "fake: config"
            __doc_py_example__ = "FakeConfig('hi': 123)"

        with pytest.raises(
            ValueError,
            match="'FakeConfig.__doc_py_example__' is not valid Python syntax",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_missing_yaml_example(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_py_example__ = "FakeConfig()"

        with pytest.raises(
            ValueError,
            match="Config model 'FakeConfig' is missing a '__doc_yaml_example__'",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_invalid_yaml_example(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_yaml_example__ = "fake: config: oops"
            __doc_py_example__ = "FakeConfig()"

        with pytest.raises(
            ValueError,
            match="'FakeConfig.__doc_yaml_example__' is not valid YAML syntax",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_non_config_does_not_have_constructors(self):
        """Models that don't end in "Config" shouldn't document constructor methods."""

        @dataclass(frozen=True)
        class FakeStatus(ModelBase):
            """FakeStatus is so cool."""

            foo: str = field(metadata={"docstring": "Foo docstring."})

            __doc_py_example__ = "FakeStatus()"
            __doc_yaml_example__ = "foo: hi"

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeStatus])]
        ).generate()

        assert "def __init__(self, **fields)" not in output["fake.md"]
        assert "def options(self, **fields)" not in output["fake.md"]
        assert "def to_dict(self) -> Dict[str, Any]" in output["fake.md"]

    def test_type_hints(self):
        @dataclass(frozen=True)
        class FakeEnum(ModelEnum):
            pass

        @dataclass(frozen=True)
        class FakeStatus(ModelBase):
            """FakeStatus is so cool."""

            __doc_py_example__ = "FakeConfig()"

        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            foo: str = field(metadata={"docstring": "foo docs"})
            bar: Optional[str] = field(metadata={"docstring": "bar docs"})
            baz: Dict[str, FakeStatus] = field(metadata={"docstring": "baz docs"})
            qux: List[Optional[FakeEnum]] = field(metadata={"docstring": "qux docs"})

            __doc_py_example__ = "FakeConfig()"
            __doc_yaml_example__ = "fake: config"

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeConfig, FakeStatus, FakeEnum])]
        ).generate()

        assert "**`foo` (str)**: foo docs" in output["fake.md"]
        assert "**`bar` (str | None)**: bar docs" in output["fake.md"]
        assert (
            "**`baz` (Dict[str, [FakeStatus](fake.md#fakestatus)])**: baz docs"
            in output["fake.md"]
        )
        assert (
            "**`qux` (List[[FakeEnum](fake.md#fakeenum) | None])**: qux docs"
            in output["fake.md"]
        )

    def test_customer_hosted_cloud_only(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does special things."""

            # Should not be labeled as customer-hosted cloud only.
            foo: str = field(metadata={"docstring": "Foo docstring."})

            __doc_yaml_example__ = """\
config: file"""

            __doc_py_example__ = """\
run_python_command()
"""

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeConfig])]
        ).generate()

        # By default options should not be labeled as customer-hosted cloud only.
        assert "Only available on [customer-hosted clouds]" not in output["fake.md"]

        @dataclass(frozen=True)
        class FakeConfigTwo(ModelBase):
            """FakeConfigTwo does really special things."""

            # Should be labeled as customer-hosted cloud only.
            foo: str = field(
                metadata={"docstring": "Foo docstring.", "customer_hosted_only": True}
            )

            __doc_yaml_example__ = """\
config: file"""

            __doc_py_example__ = """\
run_python_command()
"""

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeConfigTwo])]
        ).generate()

        assert "Only available on [customer-hosted clouds]" in output["fake.md"]


class TestGenerateModelEnumType:
    def test_basic(self):
        @dataclass(frozen=True)
        class FakeEnum(ModelEnum):
            """FakeEnum is highly important."""

            FOO = "FOO"
            BAR = "BAR"

            __docstrings__ = {
                "FOO": "Foo is quite important.",
                "BAR": "But bar is even more so.",
            }

        output = MarkdownGenerator(modules=[_fake_module(models=[FakeEnum])]).generate()
        assert len(output) == 1
        assert "FakeEnum is highly important." in output["fake.md"]
        assert "**`FOO`**: Foo is quite important." in output["fake.md"]
        assert "**`BAR`**: But bar is even more so." in output["fake.md"]

    def test_reject_value_missing_docstring(self):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "ModelEnum 'FakeEnum.__docstrings__' is missing docstrings for values: ['BAR', 'FOO']"
            ),
        ):

            @dataclass(frozen=True)
            class FakeEnum(ModelEnum):
                """FakeEnum is highly important."""

                FOO = "FOO"
                BAR = "BAR"

        with pytest.raises(
            ValueError,
            match=re.escape(
                "ModelEnum 'FakeEnumTwo.__docstrings__' is missing docstrings for values: ['BAR']"
            ),
        ):

            @dataclass(frozen=True)
            class FakeEnumTwo(ModelEnum):
                """FakeEnumTwo is also important."""

                FOO = "FOO"
                BAR = "BAR"

                __docstrings__ = {
                    "FOO": "Foo is quite important.",
                }


class TestGenerateCLICommand:
    def test_basic(self):
        @click.command("do-something", cls=AnyscaleCommand, example="infinite loop")
        @click.argument("the-arg")
        @click.option("-m", "--maybe", help="Maybe do something else.")
        def do_something():
            """This is the description of what to do."""

        output = MarkdownGenerator(
            modules=[_fake_module(cli_commands=[do_something])]
        ).generate()
        assert "## `anyscale fake do-something`" in output["fake.md"]
        assert "anyscale fake do-something [OPTIONS] THE_ARG" in output["fake.md"]
        assert "This is the description of what to do." in output["fake.md"]
        assert "**`-m/--maybe`**: Maybe do something else." in output["fake.md"]
        assert "--help" not in output["fake.md"]

    def test_secondary_opts(self):
        @click.command("do-something", cls=AnyscaleCommand, example="infinite loop")
        @click.argument("the-arg")
        @click.option("--enable/--disable", help="Enable or disable something.")
        def do_something():
            """This is the description of what to do."""

        output = MarkdownGenerator(
            modules=[_fake_module(cli_commands=[do_something])]
        ).generate()
        assert "## `anyscale fake do-something`" in output["fake.md"]
        assert "anyscale fake do-something [OPTIONS] THE_ARG" in output["fake.md"]
        assert "This is the description of what to do." in output["fake.md"]
        assert (
            "**`--enable/--disable`**: Enable or disable something."
            in output["fake.md"]
        )
        assert "--help" not in output["fake.md"]

    def test_command_with_group(self):
        @click.group("my_group", help="This is my group.")
        def my_group() -> None:
            pass

        @my_group.command("do-something", cls=AnyscaleCommand, example="infinite loop")
        def do_something():
            """This is the description of what to do."""

        output = MarkdownGenerator(
            modules=[
                _fake_module(
                    cli_commands=[do_something],
                    cli_command_group_prefix={do_something: "my_group"},
                ),
            ]
        ).generate()
        assert "## `anyscale fake my_group do-something`" in output["fake.md"]
        assert "anyscale fake my_group do-something [OPTIONS]" in output["fake.md"]
        assert "--help" not in output["fake.md"]
        assert '<TabItem value="cli" label="CLI">' in output["fake.md"]
        assert "infinite loop" in output["fake.md"]

    def test_legacy_command_with_group(self):
        @click.group("my_group", help="This is my group.")
        def my_group() -> None:
            pass

        @my_group.command("new", cls=AnyscaleCommand, example="infinite loop")
        def new():
            """This is the description of what to do."""

        @my_group.command(
            "old", cls=LegacyAnyscaleCommand, new_cli=new, new_prefix="anyscale"
        )
        def old():
            """This is the description of what to do."""

        @my_group.command("limited", cls=LegacyAnyscaleCommand, is_limited_support=True)
        def limited():
            """This is the description of what to do."""

        output = MarkdownGenerator(
            modules=[_fake_module(legacy_cli_commands=[old, limited],),]
        ).generate()
        assert "### `anyscale legacy fake old`" in output["fake.md"]
        assert '<span class="label-h3 label-legacy">Legacy</span>' in output["fake.md"]
        assert (
            "This command is deprecated. Upgrade to [anyscale new](#anyscale-new)."
            in output["fake.md"]
        )
        assert "### `anyscale legacy fake limited`" in output["fake.md"]
        assert '<span class="label-h3 label-legacy">Legacy</span>' in output["fake.md"]
        assert (
            "This command is not actively maintained. Use with caution."
            in output["fake.md"]
        )

    def test_beta_command_with_group(self):
        @click.group("my_group", help="This is my group.")
        def my_group() -> None:
            pass

        @my_group.command(
            "new", cls=AnyscaleCommand, example="infinite loop", is_beta=True
        )
        def new():
            """This is the description of what to do."""

        output = MarkdownGenerator(
            modules=[_fake_module(cli_commands=[new])]
        ).generate()
        assert "### `anyscale fake new`" in output["fake.md"]
        assert '<span class="label-h3 label-beta">Beta</span>' in output["fake.md"]
        assert (
            "This command undergoes rapid iteration. Users must be tolerant of change."
            in output["fake.md"]
        )


class TestGenerateSDKCommand:
    def test_basic(self):
        @sdk_command(
            "TEST_KEY",
            None,
            doc_py_example="do_something()",
            arg_docstrings={"s": "a string"},
        )
        def do_something(s: str, *, _sdk) -> int:
            """Might do something."""

        output = MarkdownGenerator(
            modules=[_fake_module(sdk_commands=[do_something])]
        ).generate()

        assert "## `anyscale.fake.do_something`" in output["fake.md"]
        assert "Might do something." in output["fake.md"]
        assert "**`s` (str)**: a string" in output["fake.md"]
        assert "**Returns**: int" in output["fake.md"]
        # Private arguments should not be documented.
        assert "_sdk" not in output["fake.md"]

    def test_no_return(self):
        @sdk_command(
            "TEST_KEY",
            None,
            doc_py_example="do_something()",
            arg_docstrings={"s": "a string"},
        )
        def do_something(s: str):
            """Might do something."""

        output = MarkdownGenerator(
            modules=[_fake_module(sdk_commands=[do_something])]
        ).generate()

        assert "Returns" not in output["fake.md"]

    def test_missing_docstring(self):
        @sdk_command(
            "TEST_KEY",
            None,
            doc_py_example="do_something()",
            arg_docstrings={"s": "a string"},
        )
        def do_something(s: str, *, _sdk):
            pass

        with pytest.raises(
            ValueError,
            match="SDK command 'anyscale.fake.do_something' is missing a docstring.",
        ):
            MarkdownGenerator(
                modules=[_fake_module(sdk_commands=[do_something])]
            ).generate()

    def test_missing_py_example(self):
        with pytest.raises(
            ValueError,
            match="SDK command 'do_something' must provide a non-empty 'doc_py_example'.",
        ):

            @sdk_command(
                "TEST_KEY", None, doc_py_example="", arg_docstrings={"s": "a string"},
            )
            def do_something(s: str, *, _sdk):
                """Might do something."""

    def test_missing_arg_docstring(self):
        @sdk_command(
            "TEST_KEY", None, doc_py_example="do_something()", arg_docstrings={},
        )
        def do_something(s: str, *, _sdk):
            """Might do something."""

        with pytest.raises(
            ValueError,
            match="SDK command 'anyscale.fake.do_something' is missing a docstring for argument 's'",
        ):
            MarkdownGenerator(
                modules=[_fake_module(sdk_commands=[do_something])]
            ).generate()

    def test_type_hints(self):
        """Models that don't end in "Config" shouldn't document constructor methods."""

        @dataclass(frozen=True)
        class FakeEnum(ModelEnum):
            pass

        @dataclass(frozen=True)
        class FakeStatus(ModelBase):
            """FakeStatus is so cool."""

            __doc_py_example__ = "FakeStatus()"

        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_py_example__ = "FakeConfig()"
            __doc_yaml_example__ = "fake: config"

        @sdk_command(
            "TEST_KEY",
            None,
            doc_py_example="do_something()",
            arg_docstrings={
                "foo": "foo docs",
                "bar": "bar docs",
                "baz": "baz docs",
                "qux": "qux docs",
                "quux": "quux docs",
                "quuz": "quuz docs",
            },
        )
        def do_something(
            foo: FakeEnum,
            bar: Optional[FakeStatus],
            baz: Union[Optional[FakeConfig], FakeStatus],
            qux: List[FakeEnum],
            quux: Dict[str, Optional[FakeEnum]],
            quuz: Tuple[str, int],
        ) -> Optional[FakeStatus]:
            """Might do something."""

        output = MarkdownGenerator(
            modules=[
                _fake_module(
                    sdk_commands=[do_something],
                    models=[FakeEnum, FakeStatus, FakeConfig],
                )
            ]
        ).generate()

        assert "**`foo` ([FakeEnum](fake.md#fakeenum))**: foo docs" in output["fake.md"]
        assert (
            "**`bar` ([FakeStatus](fake.md#fakestatus) | None)**: bar docs"
            in output["fake.md"]
        )
        assert (
            "**`baz` ([FakeConfig](fake.md#fakeconfig) | None | [FakeStatus](fake.md#fakestatus))**: baz docs"
            in output["fake.md"]
        )
        assert (
            "**`qux` (List[[FakeEnum](fake.md#fakeenum)])**: qux docs"
            in output["fake.md"]
        )
        assert (
            "**`quux` (Dict[str, [FakeEnum](fake.md#fakeenum) | None])**: quux docs"
            in output["fake.md"]
        )
        assert "**`quuz` (Tuple[str, int])**: quuz docs" in output["fake.md"]
        assert (
            "**Returns**: [FakeStatus](fake.md#fakestatus) | None" in output["fake.md"]
        )



import asyncio
import os
import random
import string
import time
from typing import Any, cast, Dict, List, Optional

import click
import tabulate
import yaml

from anyscale.anyscale_pydantic import BaseModel
from anyscale.api_utils.job_logs_util import _get_job_logs_from_storage_bucket_streaming
from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client import (
    ComputeTemplate,
    ProductionJob,
    ProductionJobConfig,
)
from anyscale.client.openapi_client.models.create_internal_production_job import (
    CreateInternalProductionJob,
)
from anyscale.client.openapi_client.models.create_job_queue_config import (
    CreateJobQueueConfig,
)
from anyscale.client.openapi_client.models.decorated_production_job import (
    DecoratedProductionJob,
)
from anyscale.client.openapi_client.models.ha_job_states import HaJobStates
from anyscale.controllers.base_controller import BaseController
from anyscale.models.job_model import JobConfig
from anyscale.project import infer_project_id
from anyscale.sdk.anyscale_client.models.job import Job
from anyscale.sdk.anyscale_client.models.jobs_query import JobsQuery
from anyscale.sdk.anyscale_client.models.jobs_sort_field import JobsSortField
from anyscale.sdk.anyscale_client.models.page_query import PageQuery
from anyscale.sdk.anyscale_client.models.sort_by_clause_jobs_sort_field import (
    SortByClauseJobsSortField,
)
from anyscale.sdk.anyscale_client.models.sort_order import SortOrder
from anyscale.util import (
    get_endpoint,
    is_anyscale_workspace,
    poll,
    populate_unspecified_cluster_configs_from_current_workspace,
    validate_job_config_dict,
)
from anyscale.utils.connect_helpers import search_entities
from anyscale.utils.runtime_env import override_runtime_env_config
from anyscale.utils.workload_types import Workload


log = LogsLogger()

_TERMINAL_STATES = {
    HaJobStates.SUCCESS,
    HaJobStates.TERMINATED,
    HaJobStates.BROKEN,
    HaJobStates.OUT_OF_RETRIES,
}
_PENDING_STATES = {
    HaJobStates.PENDING,
    HaJobStates.AWAITING_CLUSTER_START,
    HaJobStates.ERRORED,
    HaJobStates.RESTARTING,
}


class MiniJobRun(BaseModel):
    last_job_run_id: str
    job_state: str
    error: Optional[str]


class JobController(BaseController):
    def __init__(
        self,
        log: Optional[LogsLogger] = None,
        initialize_auth_api_client: bool = True,
        raise_structured_exception: bool = False,
        auth_token: str = None,
    ):
        if log is None:
            log = LogsLogger()

        super().__init__(
            initialize_auth_api_client, raise_structured_exception, cli_token=auth_token
        )
        self.log = log
        self.log.open_block("Output")
        self._cluster_journal_events_start_line = 0

    def submit(
        self,
        job_config_file: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_entrypoint_cmd: Optional[bool] = False,
        entrypoint: Optional[List[str]] = None,
    ) -> str:
        entrypoint = entrypoint or []
        workspace_id = os.environ.get("ANYSCALE_EXPERIMENTAL_WORKSPACE_ID", None)
        if is_anyscale_workspace() and is_entrypoint_cmd:
            entrypoint = [job_config_file, *entrypoint]
            config = self.generate_config_from_entrypoint(
                entrypoint, name, description, workspace_id
            )
            id = self.submit_from_config(config)  # noqa: A001
        elif len(entrypoint) == 0:
            # Assume that job_config_file is a file and submit it.
            config = self.generate_config_from_file(
                job_config_file,
                name=name,
                description=description,
                workspace_id=workspace_id,
            )
            id = self.submit_from_config(config)  # noqa: A001
        elif len(entrypoint) != 0:
            msg = (
                "Within an Anyscale Workspace, `anyscale job submit` takes either a file, or a command. To submit a command, use `anyscale job submit -- my command`."
                if is_anyscale_workspace()
                else "`anyscale job submit` takes one argument, a YAML file configuration. Please use `anyscale job submit my_file`."
            )
            raise click.ClickException(msg)
        return id

    def generate_config_from_entrypoint(
        self,
        entrypoint: List[str],
        name: Optional[str],
        description: Optional[str],
        workspace_id: Optional[str] = None,
    ) -> JobConfig:
        config_dict = {
            "entrypoint": " ".join(entrypoint),
            "name": name,
            "description": description,
            "workspace_id": workspace_id,
        }

        config_dict = populate_unspecified_cluster_configs_from_current_workspace(
            config_dict, self.anyscale_api_client,
        )

        return JobConfig.parse_obj(config_dict)

    def generate_config_from_file(
        self,
        job_config_file: str,
        name: Optional[str],
        description: Optional[str],
        workspace_id: Optional[str] = None,
    ) -> JobConfig:
        config_dict = self._load_config_dict_from_file(job_config_file)
        config_dict["workspace_id"] = workspace_id
        validate_job_config_dict(config_dict, self.api_client)
        config_dict = populate_unspecified_cluster_configs_from_current_workspace(
            config_dict, self.anyscale_api_client
        )
        job_config = JobConfig.parse_obj(config_dict)

        if name:
            job_config.name = name

        if description:
            job_config.description = description

        return job_config

    def submit_from_config(self, job_config: JobConfig):
        # If project id is not specified, try to infer it
        project_id = infer_project_id(
            self.anyscale_api_client,
            self.api_client,
            self.log,
            project_id=job_config.project_id,
            cluster_compute_id=job_config.compute_config_id,
            cluster_compute=job_config.compute_config,
            cloud=job_config.cloud,
        )

        job_config.runtime_env = override_runtime_env_config(
            runtime_env=job_config.runtime_env,
            anyscale_api_client=self.anyscale_api_client,
            api_client=self.api_client,
            workload_type=Workload.JOBS,
            compute_config_id=job_config.compute_config_id,
            log=self.log,
        )
        config_object = ProductionJobConfig(
            entrypoint=job_config.entrypoint,
            runtime_env=job_config.runtime_env,
            build_id=job_config.build_id,
            compute_config_id=job_config.compute_config_id,
            max_retries=job_config.max_retries,
        )

        job_queue_config = (
            CreateJobQueueConfig(**job_config.job_queue_config)
            if job_config.job_queue_config
            else None
        )

        job = self.api_client.create_job_api_v2_decorated_ha_jobs_create_post(
            CreateInternalProductionJob(
                name=job_config.name or self._generate_random_job_name(),
                description=job_config.description or "Job submitted from CLI",
                project_id=project_id,
                workspace_id=job_config.workspace_id,
                config=config_object,
                job_queue_config=job_queue_config,
            )
        ).result
        self.log.info(
            f"Maximum uptime is {self._get_maximum_uptime_output(job)} for clusters launched by this job."
        )
        self.log.info(
            f"Job {job.id} has been successfully submitted. Current state of job: {job.state.current_state}."
        )
        self.log.info(
            f"Query the status of the job with `anyscale job list --job-id {job.id}`."
        )
        self.log.info(
            f"Get the logs for the job with `anyscale job logs --job-id {job.id} --follow`."
        )
        self.log.info(f'View the job in the UI at {get_endpoint(f"/jobs/{job.id}")}')
        return job.id

    def _get_maximum_uptime_output(self, job: ProductionJob) -> str:
        compute_config: ComputeTemplate = self.api_client.get_compute_template_api_v2_compute_templates_template_id_get(
            job.config.compute_config_id
        ).result
        maximum_uptime_minutes = compute_config.config.maximum_uptime_minutes
        if maximum_uptime_minutes and maximum_uptime_minutes > 0:
            return f"set to {maximum_uptime_minutes} minutes"
        return "disabled"

    def _load_config_dict_from_file(self, job_config_file: str) -> Dict[str, Any]:
        if not os.path.exists(job_config_file):
            raise click.ClickException(f"Config file {job_config_file} not found.")

        with open(job_config_file) as f:
            config_dict = yaml.safe_load(f)
        return config_dict

    def wait(
        self,
        job_name: Optional[str] = None,
        job_id: Optional[str] = None,
        target_state: str = HaJobStates.SUCCESS,
        interval_secs: float = 10.0,
        timeout_secs=None,
    ):
        if target_state not in HaJobStates.allowable_values:
            raise click.ClickException(
                f"`{target_state}` is not a valid Job state. Allowed states are {HaJobStates.allowable_values}"
            )
        job_id_or_name = job_name or job_id
        job_id = cast(str, self._resolve_job_object(job_id, job_name).id,)
        with self.log.spinner(f"Waiting for Job `{job_id_or_name}`...") as spinner:
            state = None
            for _ in poll(interval_secs=interval_secs, timeout_secs=timeout_secs):
                state = self._get_job_state(job_id)
                spinner.text = f"Job `{job_id_or_name}` is in state `{state}`. Waiting to reach state `{target_state}`."

                if state == target_state:
                    spinner.succeed(
                        f"Job `{job_id_or_name}` reached state `{target_state}`."
                    )
                    return job_id

                if state in _TERMINAL_STATES:
                    msg = f"Job `{job_id_or_name}` reached terminal state `{state}`, and will not reach `{target_state}`."
                    spinner.fail(msg)
                    raise click.ClickException(msg)

            msg = f"Timed out after waiting for {timeout_secs} seconds. The current state is {state}."
            spinner.fail(msg)
            raise click.ClickException(msg)

    def _get_job_state(self, job_id: str):
        job = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            job_id
        ).result
        return job.state.current_state

    def list(  # noqa: PLR0913
        self,
        include_all_users: bool,
        name: Optional[str],
        job_id: Optional[str],
        project_id: Optional[str],
        include_archived: bool,
        max_items: int,
    ) -> None:
        """
        This function will list jobs.
        """
        print(f'View your Jobs in the UI at {get_endpoint("/jobs")}')

        jobs_list = []
        if job_id:
            job = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
                job_id
            ).result
            if job.is_service:
                # `job_id` belongs to a service, but this function should list jobs.
                raise click.ClickException(
                    f"ID {job_id} belongs to a Anyscale service. Please get information about "
                    f"this service with `anyscale service list --service-id {job_id}`."
                )

            output_map = {
                "Name": job.name,
                "Id": job.id,
                "Cost (dollars)": job.cost_dollars,
                "Project name": job.project.name,
                "Cluster name": job.last_job_run.cluster.name
                if job.last_job_run and job.last_job_run.cluster
                else None,
                "Current state": job.state.current_state,
                "Creator": job.creator.email,
                "Entrypoint": job.config.entrypoint
                if len(job.config.entrypoint) < 100
                else job.config.entrypoint[:100] + " ...",
            }
            output_str = "\n".join(
                [f"\t{key}: {output_map[key]}" for key in output_map]
            )
            print(output_str)
            return
        else:
            if not include_all_users:
                creator_id = (
                    self.api_client.get_user_info_api_v2_userinfo_get().result.id
                )
            else:
                creator_id = None
            resp = self.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
                project_id=project_id,
                name=name,
                creator_id=creator_id,
                type_filter="BATCH_JOB",
                archive_status="ALL" if include_archived else "NOT_ARCHIVED",
                count=10,
            )
            jobs_list.extend(resp.results)
            paging_token = resp.metadata.next_paging_token
            has_more = (paging_token is not None) and (len(jobs_list) < max_items)
            while has_more:
                resp = self.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
                    project_id=project_id,
                    name=name,
                    creator_id=creator_id,
                    type_filter="BATCH_JOB",
                    archive_status="ALL" if include_archived else "NOT_ARCHIVED",
                    count=10,
                    paging_token=paging_token,
                )
                jobs_list.extend(resp.results)
                paging_token = resp.metadata.next_paging_token
                has_more = (paging_token is not None) and (len(jobs_list) < max_items)
            jobs_list = jobs_list[:max_items]

        jobs_table = [
            [
                job.name,
                job.id,
                job.cost_dollars,
                job.project.name,
                job.last_job_run.cluster.name
                if job.last_job_run and job.last_job_run.cluster
                else None,
                job.state.current_state,
                job.creator.email,
                job.config.entrypoint
                if len(job.config.entrypoint) < 50
                else job.config.entrypoint[:50] + " ...",
            ]
            for job in jobs_list
        ]

        table = tabulate.tabulate(
            jobs_table,
            headers=[
                "NAME",
                "ID",
                "COST",
                "PROJECT NAME",
                "CLUSTER NAME",
                "CURRENT STATE",
                "CREATOR",
                "ENTRYPOINT",
            ],
            tablefmt="plain",
        )
        print(f"JOBS:\n{table}")

    def archive(self, job_id: Optional[str], job_name: Optional[str]) -> None:
        """
        This function will archive jobs.
        """
        job_resp: DecoratedProductionJob = self._resolve_job_object(job_id, job_name)
        self.api_client.archive_job_api_v2_decorated_ha_jobs_production_job_id_archive_post(
            job_resp.id
        )
        self.log.info(f"Job {job_resp.id} is successfully archived.")

    def terminate(self, job_id: Optional[str], job_name: Optional[str]) -> None:
        """
        This function will terminate jobs.
        """
        job_resp: DecoratedProductionJob = self._resolve_job_object(job_id, job_name)
        job = self.api_client.terminate_job_api_v2_decorated_ha_jobs_production_job_id_terminate_post(
            job_resp.id
        ).result
        self.log.info(f"Job {job.id} has begun terminating...")
        self.log.info(
            f" Current state of Job: {job.state.current_state}. Goal state of Job: {job.state.goal_state}"
        )
        self.log.info(
            f"Query the status of the Job with `anyscale job list --job-id {job.id}`."
        )

    def _resolve_job_object(
        self, job_id: Optional[str], job_name: Optional[str]
    ) -> DecoratedProductionJob:
        """Given job_id or job_name, retrieve decorated ha job spec"""
        if job_id is None and job_name is None:
            raise click.ClickException(
                "Either `--id` or `--name` must be passed in for Job."
            )
        if job_id:
            return self._get_job(job_id)

        jobs_list_resp: List[
            DecoratedProductionJob
        ] = self.api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
            name=job_name, type_filter="BATCH_JOB",
        ).results
        if len(jobs_list_resp) == 0:
            raise click.ClickException(
                f"No Job found with name {job_name}. Please either pass `--id` or list the "
                f"available Jobs with `anyscale job list`."
            )
        if len(jobs_list_resp) > 1:
            raise click.ClickException(
                f"Multiple Jobs found with name {job_name}. Please specify the `--id` instead."
            )
        return jobs_list_resp[0]

    def _get_job(self, job_id: str) -> DecoratedProductionJob:
        job_object = self.api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            production_job_id=job_id
        ).result
        return job_object

    def _get_formatted_latest_job_run(
        self, job: DecoratedProductionJob
    ) -> Optional[MiniJobRun]:
        job_state = job.state.current_state
        last_job_run_id = job.last_job_run_id
        if job_state in _TERMINAL_STATES and last_job_run_id is None:
            raise click.ClickException(
                f"Can't find latest job run for {job_state} job."
            )
        if not last_job_run_id:
            return None
        return MiniJobRun(
            last_job_run_id=last_job_run_id, job_state=job_state, error=job.state.error
        )

    def _wait_for_a_job_run(
        self, job_id: Optional[str], job_name: Optional[str]
    ) -> DecoratedProductionJob:
        """Waits until the job has a job run and is in a non-pending state
        (encountered when initially submitting).

        Returns the job."""
        start = time.monotonic()
        job = self._resolve_job_object(job_id, job_name)
        last_job_run_id: str = job.last_job_run_id
        job_state = job.state.current_state
        with self.log.spinner("Waiting for job run...") as spinner:
            while last_job_run_id is None or job_state in _PENDING_STATES:
                if job_state in _TERMINAL_STATES:
                    raise click.ClickException(
                        f"Can't find latest job run for {job_state} job."
                    )
                if self._cluster_journal_events_start_line == 0:
                    spinner.text = (
                        f"Waiting for a job run, current state is {job_state}..."
                    )
                time.sleep(max(0, 3 - (time.monotonic() - start)))
                start = time.monotonic()
                job = self._get_job(job.id)
                last_job_run_id = job.last_job_run_id
                if job.state.cluster_id:
                    # Once we have a cluster id, we start printing the cluster journal event logs with each loop.
                    cluster_journal_events = self._api_client.get_startup_logs_api_v2_sessions_session_id_startup_logs_get(
                        job.state.cluster_id,
                        start_line=self._cluster_journal_events_start_line,
                        end_line=10000000,
                    ).result
                    lines = cluster_journal_events.lines.splitlines()
                    if len(lines):
                        spinner.succeed("Job run found. Cluster launching...")
                    self._cluster_journal_events_start_line = (
                        cluster_journal_events.start_line
                        + cluster_journal_events.num_lines
                    )

                    for line in lines:
                        self.log.info(line)

                job_state = job.state.current_state
        return job

    def _print_logs_using_streaming_job_logs(
        self,
        job_run_id: str,
        follow: bool = False,
        no_cluster_journal_events: bool = False,
    ) -> None:
        """Retrieves logs from the streaming job logs folder in S3/GCS"""
        logs = ""
        next_page_token = None
        # Keep track of if any logs have been printed yet,
        # If not, we assume the cluster recently launched and we
        # give more leeway for the logs to be available.
        have_jobs_started_producing_logs = False
        with self.log.spinner("Retrieving logs..."):
            (
                logs,
                next_page_token,
                cluster_journal_events_start_line,
                job_finished,
            ) = asyncio.run(
                _get_job_logs_from_storage_bucket_streaming(
                    self._api_client,
                    self.log,
                    job_run_id=job_run_id,
                    remove_escape_chars=False,
                    next_page_token=next_page_token,
                    cluster_journal_events_start_line=self._cluster_journal_events_start_line,
                    no_cluster_journal_events=no_cluster_journal_events,
                )
            )

        if len(logs):
            have_jobs_started_producing_logs = True

        for line in logs.splitlines():
            self.log.log(line)

        # Continuously poll for logs.
        terminal_state_count = 0
        while follow:
            start = time.monotonic()
            if job_finished:
                # Let's wait before terminating the loop to give the logs a chance to catch up.
                terminal_state_count += 1
                # Once jobs produce logs, we know we do not have to wait as long for those logs
                # to be uploaded to S3 since the cluster is already running.
                max_iterations = 4 if have_jobs_started_producing_logs else 6
                if terminal_state_count >= max_iterations:
                    self.log.info("Job run reached terminal state.")
                    return

            (
                logs,
                next_page_token,
                cluster_journal_events_start_line,
                job_finished,
            ) = asyncio.run(
                _get_job_logs_from_storage_bucket_streaming(
                    self._api_client,
                    self.log,
                    job_run_id=job_run_id,
                    remove_escape_chars=False,
                    next_page_token=next_page_token,
                    cluster_journal_events_start_line=cluster_journal_events_start_line,
                    no_cluster_journal_events=no_cluster_journal_events,
                )
            )

            for line in logs.splitlines():
                self.log.log(line)

            # Wait at least 5 seconds between iterations.
            time.sleep(max(0, 5 - (time.monotonic() - start)))

    def logs(
        self,
        job_id: Optional[str] = None,
        job_name: Optional[str] = None,
        should_follow: bool = False,
        all_attempts: bool = False,
        no_cluster_journal_events: bool = False,
    ) -> None:
        """
        Fetches logs for a production job using the streaming log S3 bucket.
        Params:
        - no_cluster_journal_events: Controls whether we print out cluster journal logs. (Normally printed to stderr)
        """
        job = self._wait_for_a_job_run(job_id, job_name)
        last_job_run_id: str = job.last_job_run_id
        job_run_ids: List[str] = (
            [last_job_run_id]
            if not all_attempts
            else [job_run.id for job_run in self._get_all_job_runs(job.id)]
        )  # type: ignore
        for job_run_id in job_run_ids:
            self.log.open_block(
                job_run_id, f"Job Run Id: {job_run_id}", auto_close=True,
            )
            self._print_logs_using_streaming_job_logs(
                job_run_id,
                follow=should_follow and job_run_id == last_job_run_id,
                no_cluster_journal_events=no_cluster_journal_events,
            )

    def _generate_random_job_name(self) -> str:
        """Generates a random job name
        Format:
            cli-job-{10 random characters}
        """
        random_chars = "".join(random.choices(string.ascii_letters, k=10)).lower()
        return "cli-job-" + random_chars

    def _get_all_job_runs(self, job_id: str) -> List[Job]:
        """Returns all job runs for a given job id.
        Returned in ascending order by creation time."""
        job_runs: List[Job] = search_entities(
            self.anyscale_api_client.search_jobs,
            JobsQuery(
                ha_job_id=job_id,
                show_ray_client_runs_only=False,
                sort_by_clauses=[
                    SortByClauseJobsSortField(
                        sort_field=JobsSortField.CREATED_AT, sort_order=SortOrder.ASC,
                    )
                ],
                paging=PageQuery(),
            ),
        )
        return job_runs



"""Entrypoint to generate reference docs for the Anyscale CLI/SDK.

Usage: python -m anyscale._private.docgen --help
"""
import os

import click

import anyscale
from anyscale import scripts
from anyscale._private.docgen.generator import MarkdownGenerator, Module
from anyscale.commands import (
    cloud_commands,
    cluster_commands,
    cluster_env_commands,
    compute_config_commands,
    image_commands,
    job_commands,
    login_commands,
    logs_commands,
    machine_commands,
    machine_pool_commands,
    project_commands,
    resource_quota_commands,
    schedule_commands,
    service_account_commands,
    service_commands,
    workspace_commands,
    workspace_commands_v2,
)
from anyscale.commands.llm import dataset_commands, models_commands
from anyscale.compute_config.models import (
    CloudDeployment,
    ComputeConfig,
    ComputeConfigVersion,
    HeadNodeConfig,
    MarketType,
    WorkerNodeGroupConfig,
)
from anyscale.image.models import ImageBuild, ImageBuildStatus
from anyscale.job.models import (
    JobConfig,
    JobLogMode,
    JobQueueConfig,
    JobQueueExecutionMode,
    JobQueueSpec,
    JobRunState,
    JobRunStatus,
    JobState,
    JobStatus,
)
from anyscale.llm.dataset._private.models import Dataset
from anyscale.llm.model.models import (
    DeletedFineTunedModel,
    FineTunedModel,
    FineTuningType,
)
from anyscale.schedule.models import ScheduleConfig, ScheduleState, ScheduleStatus
from anyscale.service.models import (
    RayGCSExternalStorageConfig,
    ServiceConfig,
    ServiceState,
    ServiceStatus,
    ServiceVersionState,
    ServiceVersionStatus,
    TracingConfig,
)
from anyscale.workspace.models import WorkspaceConfig


# Defines all modules to be documented.
ALL_MODULES = [
    Module(
        title="Job",
        filename="job-api.md",
        cli_prefix="anyscale job",
        cli_commands=[
            job_commands.submit,
            job_commands.status,
            job_commands.terminate,
            job_commands.archive,
            job_commands.logs,
            job_commands.wait,
            job_commands.list,
        ],
        sdk_prefix="anyscale.job",
        sdk_commands=[
            anyscale.job.submit,
            anyscale.job.status,
            anyscale.job.terminate,
            anyscale.job.archive,
            anyscale.job.get_logs,
            anyscale.job.wait,
        ],
        models=[
            JobQueueSpec,
            JobQueueConfig,
            JobQueueExecutionMode,
            JobConfig,
            JobState,
            JobStatus,
            JobRunStatus,
            JobRunState,
            JobLogMode,
        ],
        # The following commands are legacy
        legacy_sdk_commands={
            "create_job": anyscale.job.submit,
            "get_production_job": anyscale.job.status,
            "get_session_for_job": anyscale.job.status,
            "terminate_job": anyscale.job.terminate,
            "fetch_job_logs": anyscale.job.get_logs,
            "fetch_production_job_logs": anyscale.job.get_logs,
            "get_job_logs_download": anyscale.job.get_logs,
            "get_job_logs_stream": anyscale.job.get_logs,
            # limited support, no replacement yet
            "search_jobs": None,
            "list_production_jobs": None,
        },
        legacy_sdk_models=[
            "BaseJobStatus",
            "CreateClusterComputeConfig",
            "CreateJobQueueConfig",
            "CreateProductionJob",
            "CreateProductionJobConfig",
            "HaJobGoalStates",
            "HaJobStates",
            "Job",
            "JobListResponse",
            "JobQueueConfig",
            "JobQueueExecutionMode",
            "JobQueueSpec",
            "JobRunType",
            "JobStatus",
            "JobsSortField",
            "ProductionJob",
            "ProductionJobConfig",
            "ProductionJobStateTransition",
            "ProductionjobListResponse",
            "ProductionjobResponse",
            "RayRuntimeEnvConfig",
            "SortByClauseJobsSortField",
        ],
    ),
    Module(
        title="Schedule",
        filename="schedule-api.md",
        cli_prefix="anyscale schedule",
        cli_commands=[
            schedule_commands.apply,
            schedule_commands.list,
            schedule_commands.pause,
            schedule_commands.resume,
            schedule_commands.status,
            schedule_commands.trigger,
            schedule_commands.url,
        ],
        sdk_prefix="anyscale.schedule",
        sdk_commands=[
            anyscale.schedule.apply,
            anyscale.schedule.set_state,
            anyscale.schedule.status,
            anyscale.schedule.trigger,
        ],
        models=[ScheduleConfig, ScheduleState, ScheduleStatus],
        # The following commands are legacy
        legacy_sdk_commands={
            "create_or_update_schedule": anyscale.schedule.apply,
            "get_schedule": anyscale.schedule.status,
            "pause_schedule": anyscale.schedule.set_state,
            "run_schedule": anyscale.schedule.trigger,
            # limited support, no replacement yet
            "list_schedules": None,
        },
        legacy_cli_prefix="anyscale schedule",
        legacy_cli_commands={schedule_commands.create, schedule_commands.update,},
    ),
    Module(
        title="Service",
        filename="service-api.md",
        cli_prefix="anyscale service",
        cli_commands=[
            service_commands.deploy,
            service_commands.status,
            service_commands.wait,
            service_commands.rollback,
            service_commands.terminate,
        ],
        sdk_prefix="anyscale.service",
        sdk_commands=[
            anyscale.service.deploy,
            anyscale.service.status,
            anyscale.service.wait,
            anyscale.service.rollback,
            anyscale.service.terminate,
        ],
        models=[
            ServiceConfig,
            RayGCSExternalStorageConfig,
            TracingConfig,
            ServiceStatus,
            ServiceState,
            ServiceVersionStatus,
            ServiceVersionState,
        ],
        # The following commands are legacy
        legacy_sdk_commands={
            "get_service": anyscale.service.status,
            "rollback_service": anyscale.service.rollback,
            "rollout_service": anyscale.service.deploy,
            "terminate_service": anyscale.service.terminate,
            # limited support, no replacement yet
            "list_services": None,
        },
        legacy_sdk_models=[
            "ApplyProductionServiceV2Model",
            "ApplyServiceModel",
            "ListServiceModel",
            "ListservicemodelListResponse",
            "ProductionServiceV2Model",
            "ProductionServiceV2VersionModel",
            "Productionservicev2ModelResponse",
            "RollbackServiceModel",
            "RolloutStrategy",
            "ServiceConfig",
            "ServiceEventCurrentState",
            "ServiceGoalStates",
            "ServiceModel",
            "ServiceObservabilityUrls",
            "ServiceSortField",
            "ServiceType",
            "ServiceVersionState",
            "ServicemodelListResponse",
            "ServicemodelResponse",
        ],
        legacy_cli_prefix="anyscale service",
        legacy_cli_commands={service_commands.rollout},
    ),
    Module(
        title="Compute Config",
        filename="compute-config-api.md",
        cli_prefix="anyscale compute-config",
        cli_commands=[
            compute_config_commands.create_compute_config,
            compute_config_commands.get_compute_config,
            compute_config_commands.archive_compute_config,
        ],
        sdk_prefix="anyscale.compute_config",
        sdk_commands=[
            anyscale.compute_config.create,
            anyscale.compute_config.get,
            anyscale.compute_config.archive,
        ],
        models=[
            ComputeConfig,
            HeadNodeConfig,
            WorkerNodeGroupConfig,
            MarketType,
            CloudDeployment,
            ComputeConfigVersion,
        ],
        legacy_sdk_commands={
            "create_cluster_compute": anyscale.compute_config.create,
            "delete_cluster_compute": anyscale.compute_config.archive,
            "get_cluster_compute": anyscale.compute_config.get,
            "get_default_cluster_compute": anyscale.compute_config.get,
            # limited support, no replacement yet
            "search_cluster_computes": None,
        },
        legacy_sdk_models=[
            "ClusterCompute",
            "ClusterComputeConfig",
            "ClusterComputesQuery",
            "ClustercomputeListResponse",
            "ClustercomputeResponse",
            "ComputeNodeType",
            "ComputeTemplate",
            "ComputeTemplateConfig",
            "ComputeTemplateQuery",
            "ComputetemplateListResponse",
            "ComputetemplateResponse",
            "ComputetemplateconfigResponse",
            "CreateClusterCompute",
            "CreateClusterComputeConfig",
            "CreateComputeTemplate",
            "CreateComputeTemplateConfig",
            "UpdateComputeTemplate",
            "UpdateComputeTemplateConfig",
        ],
    ),
    Module(
        title="Service Account",
        filename="service-account.md",
        cli_prefix="anyscale service-account",
        cli_commands=[
            service_account_commands.create,
            service_account_commands.create_api_key,
            service_account_commands.list_service_accounts,
            service_account_commands.delete,
            service_account_commands.rotate_api_keys,
        ],
        sdk_prefix="anyscale.service_account",
        sdk_commands=[],
        models=[],
    ),
    Module(
        title="Image",
        filename="image.md",
        cli_prefix="anyscale image",
        cli_commands=[
            image_commands.build,
            image_commands.get,
            image_commands.register,
        ],
        sdk_prefix="anyscale.image",
        sdk_commands=[
            anyscale.image.build,
            anyscale.image.get,
            anyscale.image.register,
        ],
        models=[ImageBuild, ImageBuildStatus],
        legacy_title="Cluster environment",
        legacy_cli_prefix="anyscale image",
        legacy_cli_commands={
            cluster_env_commands.archive,
            cluster_env_commands.build,
            cluster_env_commands.get,
            cluster_env_commands.list,
        },
        legacy_sdk_commands={
            "get_cluster_environment_build_operation": None,
            "create_byod_cluster_environment_build": anyscale.image.build,
            "create_cluster_environment_build": anyscale.image.build,
            "find_cluster_environment_build_by_identifier": None,
            "get_cluster_environment_build": None,
            "get_cluster_environment_build_logs": None,
            "get_default_cluster_environment_build": None,
            "list_cluster_environment_builds": None,
            "create_byod_cluster_environment": anyscale.image.build,
            "create_cluster_environment": anyscale.image.build,
            "delete_cluster_environment": None,
            "get_cluster_environment": None,
            "search_cluster_environments": None,
        },
        legacy_sdk_models=[
            "ClusterEnvironment",
            "ClusterEnvironmentBuild",
            "ClusterEnvironmentBuildLogResponse",
            "ClusterEnvironmentBuildOperation",
            "ClusterEnvironmentBuildStatus",
            "ClusterEnvironmentsQuery",
            "ClusterenvironmentListResponse",
            "ClusterenvironmentResponse",
            "ClusterenvironmentbuildListResponse",
            "ClusterenvironmentbuildlogresponseResponse",
            "ClusterenvironmentbuildoperationResponse",
            "CreateBYODClusterEnvironment",
            "CreateBYODClusterEnvironmentBuild",
            "CreateBYODClusterEnvironmentConfigurationSchema",
            "CreateClusterEnvironment",
            "CreateClusterEnvironmentBuild",
            "CreateClusterEnvironmentConfigurationSchema",
            "RayRuntimeEnvConfig",
            "RuntimeEnvironment",
            "RuntimeenvironmentResponse",
        ],
    ),
    Module(
        title="Cloud",
        filename="cloud.md",
        cli_prefix="anyscale cloud",
        cli_commands=[
            cloud_commands.setup_cloud,
            cloud_commands.register_cloud,
            cloud_commands.cloud_edit,
            cloud_commands.cloud_update,
            cloud_commands.cloud_delete,
            cloud_commands.cloud_verify,
            cloud_commands.list_cloud,
            cloud_commands.cloud_config_update,
        ],
        sdk_prefix="anyscale.cloud",
        sdk_commands=[],
        models=[],
        cli_command_group_prefix={cloud_commands.cloud_config_update: "config"},
        legacy_sdk_commands={
            # limited support, no replacement yet
            "create_cloud": None,
            "delete_cloud": None,
            "get_cloud": None,
            "get_default_cloud": None,
            "search_clouds": None,
            "update_cloud": None,
        },
        legacy_sdk_models=[
            "Cloud",
            "CloudConfig",
            "CloudListResponse",
            "CloudProviders",
            "CloudResponse",
            "CloudState",
            "CloudStatus",
            "CloudTypes",
            "CloudVersion",
            "CloudsQuery",
            "CreateCloud",
            "UpdateCloud",
        ],
    ),
    Module(
        title="Logs",
        filename="logs.md",
        cli_prefix="anyscale logs",
        cli_commands=[logs_commands.anyscale_logs_cluster],
        sdk_prefix="anyscale.logs",
        sdk_commands=[],
        models=[],
    ),
    Module(
        title="Workspaces",
        filename="workspaces.md",
        cli_prefix="anyscale workspace_v2",
        cli_commands=[
            workspace_commands_v2.create,
            workspace_commands_v2.start,
            workspace_commands_v2.terminate,
            workspace_commands_v2.status,
            workspace_commands_v2.wait,
            workspace_commands_v2.ssh,
            workspace_commands_v2.run_command,
            workspace_commands_v2.pull,
            workspace_commands_v2.push,
        ],
        sdk_prefix="anyscale.workspace",
        sdk_commands=[
            anyscale.workspace.create,
            anyscale.workspace.start,
            anyscale.workspace.terminate,
            anyscale.workspace.status,
            anyscale.workspace.wait,
            anyscale.workspace.generate_ssh_config_file,
            anyscale.workspace.run_command,
        ],
        models=[WorkspaceConfig],
        legacy_cli_prefix="anyscale workspace",
        legacy_cli_commands=[
            workspace_commands.create,
            workspace_commands.run,
            workspace_commands.ssh,
            workspace_commands.start,
            workspace_commands.terminate,
            workspace_commands.pull,
            workspace_commands.push,
            # limited support, no replacement yet
            workspace_commands.clone,
            workspace_commands.copy_command,
        ],
    ),
    Module(
        title="Machine Pool",
        filename="machine-pool.md",
        cli_prefix="anyscale machine-pool",
        cli_commands=[
            machine_pool_commands.create_machine_pool,
            machine_pool_commands.delete_machine_pool,
            machine_pool_commands.list_machine_pools,
        ],
        sdk_prefix="anyscale.machine_pool",
        sdk_commands=[],
        models=[],
    ),
    Module(
        title="Cluster",
        filename="cluster.md",
        cli_prefix="anyscale cluster",
        cli_commands=[],
        sdk_prefix="anyscale.cluster",  # fake name used as id, cluster has no SDK commands
        sdk_commands=[],
        models=[],
        legacy_cli_commands=[cluster_commands.start, cluster_commands.terminate,],
        legacy_cli_prefix="anyscale cluster",
        legacy_sdk_commands={
            "launch_cluster": None,
            "launch_cluster_with_new_cluster_environment": None,
            "archive_cluster": None,
            "create_cluster": None,
            "delete_cluster": None,
            "get_cluster": None,
            "search_clusters": None,
            "start_cluster": None,
            "terminate_cluster": None,
            "update_cluster": None,
        },
        legacy_sdk_models=[
            "ClusterHeadNodeInfo",
            "ClusterListResponse",
            "ClusterManagementStackVersions",
            "ClusterOperation",
            "ClusterResponse",
            "ClusterServicesUrls",
            "ClusterState",
            "ClusterStatus",
            "ClusterStatusDetails",
            "ClusteroperationResponse",
            "ClustersQuery",
            "CreateCluster",
            "StartClusterOptions",
            "TerminateClusterOptions",
            "UpdateCluster",
        ],
    ),
    Module(
        title="LLM Model",
        filename="llm_models.md",
        cli_prefix="anyscale llm model",
        cli_commands=[
            models_commands.get_model,
            models_commands.list_models,
            models_commands.delete_model,
        ],
        sdk_prefix="anyscale.llm.model",
        sdk_commands=[
            anyscale.llm.model.get,
            anyscale.llm.model.list,
            anyscale.llm.model.delete,
        ],
        models=[FineTunedModel, FineTuningType, DeletedFineTunedModel],
    ),
    Module(
        title="LLM Dataset",
        filename="llm_dataset.md",
        cli_prefix="anyscale llm dataset",
        cli_commands=[
            dataset_commands.get_dataset,
            dataset_commands.upload_dataset,
            dataset_commands.download_dataset,
        ],
        sdk_prefix="anyscale.llm.dataset",
        sdk_commands=[
            anyscale.llm.dataset.get,
            anyscale.llm.dataset.upload,
            anyscale.llm.dataset.download,
        ],
        models=[Dataset],
    ),
    Module(
        title="Resource quotas",
        filename="resource-quotas.md",
        cli_prefix="anyscale resource-quota",
        cli_commands=[
            resource_quota_commands.create,
            resource_quota_commands.list_resource_quotas,
            resource_quota_commands.delete,
            resource_quota_commands.enable,
            resource_quota_commands.disable,
        ],
        sdk_prefix="anyscale.resource_quota",
        sdk_commands=[],  # No SDK commands for resource quotas
        models=[],
    ),
    Module(
        title="Other",
        filename="other.md",
        cli_prefix="anyscale",
        cli_commands=[
            login_commands.anyscale_login,
            login_commands.anyscale_logout,
            scripts.version_cli,
        ],
        sdk_prefix="",
        sdk_commands=[],
        models=[],
        # The following commands are legacy
        legacy_cli_prefix="anyscale",
        legacy_cli_commands=[
            machine_commands.delete_machine,
            machine_commands.list_machines,
            project_commands.create,
            project_commands.list,
        ],
        legacy_sdk_commands={
            "get_organization_temporary_object_storage_credentials": None,
            "partial_update_organization": None,
            "create_project": None,
            "delete_project": None,
            "get_default_project": None,
            "get_project": None,
            "search_projects": None,
            "update_project": None,
            "get_runtime_environment": None,
            "create_session_command": None,
            "get_session_command": None,
            "list_session_commands": None,
            "get_session_event_log": None,
            "upsert_sso_config": None,
            "upsert_test_sso_config": None,
        },
        legacy_sdk_models=[
            "AWSNodeOptions",
            "AWSTag",
            "AWSTagSpecification",
            "AccessConfig",
            "ArchiveStatus",
            "BASEIMAGESENUM",
            "BaseJobStatus",
            "BlockDeviceMapping",
            "Build",
            "BuildListResponse",
            "BuildResponse",
            "BuildStatus",
            "CreateBuild",
            "CreateProject",
            "CreateSSOConfig",
            "EbsBlockDevice",
            "GCPNodeDisk",
            "GCPNodeOptions",
            "GrpcProtocolConfig",
            "HTTPValidationError",
            "HttpProtocolConfig",
            "IamInstanceProfileSpecification",
            "IdleTerminationStatus",
            "ListResponseMetadata",
            "NetworkInterface",
            "NodeType",
            "ObjectStorageConfig",
            "ObjectStorageConfigS3",
            "ObjectstorageconfigResponse",
            "OperationError",
            "OperationProgress",
            "OperationResult",
            "Organization",
            "OrganizationResponse",
            "PageQuery",
            "Project",
            "ProjectListResponse",
            "ProjectResponse",
            "ProjectsQuery",
            "Protocols",
            "PythonModules",
            "PythonVersion",
            "RayGCSExternalStorageConfig",
            "Resources",
            "SSOConfig",
            "SSOMode",
            "SUPPORTEDBASEIMAGESENUM",
            "SortOrder",
            "SsoconfigResponse",
            "StaticSSOConfig",
            "TextQuery",
            "TracingConfig",
            "UpdateOrganization",
            "UpdateProject",
            "UserServiceAccessTypes",
            "WorkerNodeType",
        ],
    ),
]


@click.command(help="Generate markdown docs for the Anyscale CLI & SDK.")
@click.argument("output_dir")
@click.option(
    "-r",
    "--remove-existing",
    is_flag=True,
    default=False,
    help="If set, all files in the 'output_dir' that were not generated will be removed.",
)
def generate(
    output_dir: str, *, remove_existing: bool = False,
):
    if not os.path.isdir(output_dir):
        raise RuntimeError(f"output_dir '{output_dir}' does not exist.")

    gen = MarkdownGenerator(ALL_MODULES)
    gen.generate()

    generated_files = set()
    os.makedirs(output_dir, exist_ok=True)
    for filename, content in gen.generate().items():
        generated_files.add(filename)
        full_path = os.path.join(output_dir, filename)
        print(f"Writing output file {full_path}")
        with open(full_path, "w") as f:
            f.write(content)

    if remove_existing:
        to_remove = set(os.listdir(output_dir)) - generated_files
        for path in to_remove:
            full_path = os.path.join(output_dir, path)
            print(f"Removing existing file {full_path}")
            os.unlink(full_path)


if __name__ == "__main__":
    generate()



import hashlib
from tempfile import NamedTemporaryFile, TemporaryDirectory
from unittest.mock import Mock, patch

from click.testing import CliRunner

import anyscale
from anyscale.commands.llm.dataset_commands import download_dataset


def test_download():
    runner = CliRunner()

    # 1. Errors if file already exists
    with NamedTemporaryFile() as f:
        # anyscale llm dataset download my_dataset -o /tmp/existing_file
        result = runner.invoke(download_dataset, args=["my_dataset", "-o", f.name])
        assert result.exception
        assert "File already exists" in result.output

    # 2. Downloads the dataset to the specified file
    unique_str = hashlib.md5().hexdigest()
    with patch.object(
        anyscale.llm.dataset,
        anyscale.llm.dataset.download.__name__,
        new=Mock(return_value=unique_str.encode()),
    ), TemporaryDirectory() as tmp_dir:
        file_path = tmp_dir + "/my_dataset"
        result = runner.invoke(download_dataset, args=["my_dataset", "-o", file_path])
        assert "Dataset 'my_dataset' downloaded to" in result.output
        with open(file_path, "r") as output_file:
            assert output_file.read() == unique_str



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class CreateJobQueueConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'priority': 'int',
        'target_job_queue_id': 'str',
        'target_job_queue_name': 'str',
        'job_queue_spec': 'JobQueueSpec'
    }

    attribute_map = {
        'priority': 'priority',
        'target_job_queue_id': 'target_job_queue_id',
        'target_job_queue_name': 'target_job_queue_name',
        'job_queue_spec': 'job_queue_spec'
    }

    def __init__(self, priority=None, target_job_queue_id=None, target_job_queue_name=None, job_queue_spec=None, local_vars_configuration=None):  # noqa: E501
        """CreateJobQueueConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._priority = None
        self._target_job_queue_id = None
        self._target_job_queue_name = None
        self._job_queue_spec = None
        self.discriminator = None

        if priority is not None:
            self.priority = priority
        if target_job_queue_id is not None:
            self.target_job_queue_id = target_job_queue_id
        if target_job_queue_name is not None:
            self.target_job_queue_name = target_job_queue_name
        if job_queue_spec is not None:
            self.job_queue_spec = job_queue_spec

    @property
    def priority(self):
        """Gets the priority of this CreateJobQueueConfig.  # noqa: E501

        Job's relative priority (only relevant for Job Queues of type PRIORITY). Valid values range from 0 (highest) to +inf (lowest). Default value is None  # noqa: E501

        :return: The priority of this CreateJobQueueConfig.  # noqa: E501
        :rtype: int
        """
        return self._priority

    @priority.setter
    def priority(self, priority):
        """Sets the priority of this CreateJobQueueConfig.

        Job's relative priority (only relevant for Job Queues of type PRIORITY). Valid values range from 0 (highest) to +inf (lowest). Default value is None  # noqa: E501

        :param priority: The priority of this CreateJobQueueConfig.  # noqa: E501
        :type: int
        """

        self._priority = priority

    @property
    def target_job_queue_id(self):
        """Gets the target_job_queue_id of this CreateJobQueueConfig.  # noqa: E501

        Identifier of the existing Job Queue this job should be added to. Note, only one of `target_job_queue_id`, `target_job_queue_name` or `job_queue_spec` could be provided  # noqa: E501

        :return: The target_job_queue_id of this CreateJobQueueConfig.  # noqa: E501
        :rtype: str
        """
        return self._target_job_queue_id

    @target_job_queue_id.setter
    def target_job_queue_id(self, target_job_queue_id):
        """Sets the target_job_queue_id of this CreateJobQueueConfig.

        Identifier of the existing Job Queue this job should be added to. Note, only one of `target_job_queue_id`, `target_job_queue_name` or `job_queue_spec` could be provided  # noqa: E501

        :param target_job_queue_id: The target_job_queue_id of this CreateJobQueueConfig.  # noqa: E501
        :type: str
        """

        self._target_job_queue_id = target_job_queue_id

    @property
    def target_job_queue_name(self):
        """Gets the target_job_queue_name of this CreateJobQueueConfig.  # noqa: E501

        Existing Job Queue user-provided name (identifier), this job should be added to. Note, only one of `target_job_queue_id`, `target_job_queue_name` or `job_queue_spec` could be provided  # noqa: E501

        :return: The target_job_queue_name of this CreateJobQueueConfig.  # noqa: E501
        :rtype: str
        """
        return self._target_job_queue_name

    @target_job_queue_name.setter
    def target_job_queue_name(self, target_job_queue_name):
        """Sets the target_job_queue_name of this CreateJobQueueConfig.

        Existing Job Queue user-provided name (identifier), this job should be added to. Note, only one of `target_job_queue_id`, `target_job_queue_name` or `job_queue_spec` could be provided  # noqa: E501

        :param target_job_queue_name: The target_job_queue_name of this CreateJobQueueConfig.  # noqa: E501
        :type: str
        """

        self._target_job_queue_name = target_job_queue_name

    @property
    def job_queue_spec(self):
        """Gets the job_queue_spec of this CreateJobQueueConfig.  # noqa: E501

        Spec of the Job Queue definition that should be created and associated with this job. Note, only one of `target_job_queue_id`, `target_job_queue_name` or `job_queue_spec` could be provided  # noqa: E501

        :return: The job_queue_spec of this CreateJobQueueConfig.  # noqa: E501
        :rtype: JobQueueSpec
        """
        return self._job_queue_spec

    @job_queue_spec.setter
    def job_queue_spec(self, job_queue_spec):
        """Sets the job_queue_spec of this CreateJobQueueConfig.

        Spec of the Job Queue definition that should be created and associated with this job. Note, only one of `target_job_queue_id`, `target_job_queue_name` or `job_queue_spec` could be provided  # noqa: E501

        :param job_queue_spec: The job_queue_spec of this CreateJobQueueConfig.  # noqa: E501
        :type: JobQueueSpec
        """

        self._job_queue_spec = job_queue_spec

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateJobQueueConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateJobQueueConfig):
            return True

        return self.to_dict() != other.to_dict()



import click


def try_import_ray():
    try:
        import ray

        return ray
    except ImportError:
        raise click.ClickException(
            "Ray not installed locally on this machine but required "
            "for the command. Please install with `pip install 'anyscale[all]'`."
        )



from io import FileIO
import os

from rich.progress import (
    BarColumn,
    FileSizeColumn,
    Progress,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TotalFileSizeColumn,
)
from rich.style import Style
from rich.table import Column

from anyscale.shared_anyscale_utils.bytes_util import Bytes


class FileDownloadProgress(Progress):
    """
    Returns a progress table for SDK logs / CLI output that looks like:

    ```
    0:00:01  50% [=========>---------]  1.5 MB / 3 MB  Downloading 'file.txt'
    ```

    (`Downloading 'file.txt'` is the task description)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            # Columns
            TimeElapsedColumn(),  #     1. Time elapsed
            TaskProgressColumn(),  #    2. Percent complete
            BarColumn(),  #             3. Horizontal progress bar
            FileSizeColumn(),  #        4. {uploaded}/{total} filesize
            "/",
            TotalFileSizeColumn(),
            TextColumn(  #              5. Current step
                "{task.description}",
                table_column=Column(width=50),
                style=Style(bold=True, color="blue"),
            ),
            *args,
            **kwargs
        )


class ProgressFileReader:
    """
    A file reader that updates a progress table (ie. the number of bytes read) as it reads the file.

    This is required because:
    - S3 presigned upload URLs upload the file in one go
    - Thus, we use the file read position as a proxy for the upload progress
    """

    def __init__(
        self, file: FileIO, progress: Progress, task_id: TaskID,
    ):
        self.file = file
        self.progress = progress
        self.task_id = task_id
        self.bytes_read = 0
        self.file_size = os.path.getsize(file.name)

    def read(self, size=-1):
        chunk = self.file.read(size)
        if chunk:
            self.bytes_read += len(chunk)
            self.progress.update(self.task_id, advance=len(chunk))
        return chunk

    def __iter__(self):
        return self

    def __next__(self):
        chunk = self.read(Bytes.KB)
        if not chunk:
            raise StopIteration
        return chunk

    def __len__(self):
        return self.file_size



import re
from typing import List, Tuple


ANYSCALE_SDK_INTRO = """\
The AnyscaleSDK class must be constructed in order to make calls to the SDK. This class allows you to create an authenticated client in which to use the SDK.

| Param | Type | Description |
| :--- | :--- | :--- |
| `auth_token` | Optional String | Authentication token used to verify you have permissions to access Anyscale. If not provided, permissions default to the credentials set for your current user. Credentials can be set by following the instructions on this page: https://console.anyscale.com/credentials |

**Example**
```python
from anyscale import AnyscaleSDK

sdk = AnyscaleSDK()
```
"""


class LegacySDK:
    def __init__(self, name: str, docstring: str):
        self.name = name
        self.docstring = docstring

    @classmethod
    def from_md(cls, md: str) -> "LegacySDK":
        """
        Convert a blob of markdown into a LegacySDK object.
        """
        name = ""
        docstring = ""

        for line in md.split("\n"):
            if line.startswith("### "):
                name = line[4:]
            else:
                docstring += (
                    re.sub("\\(./models.md#([a-z]+)\\)", "(#\\1-legacy)", line) + "\n"
                )

        return cls(name=name, docstring=docstring.strip())


class LegacyModel:
    def __init__(self, name: str, docstring: str):
        self.name = name
        self.docstring = docstring

    @classmethod
    def from_md(cls, md: str) -> "LegacyModel":
        """
        Convert a blob of markdown into a LegacySDK object.
        """
        name = ""
        docstring = ""

        for line in md.split("\n"):
            if line.startswith("## "):
                name = line[3:]
            else:
                docstring += re.sub("\\(#([a-z]+)\\)", "(#\\1-legacy)", line) + "\n"

        return cls(name=name, docstring=docstring.strip())


def _chunk(md_file: str, start: str, ends: List[str]) -> List[str]:
    """
    Split a markdown file into chunks based on the header.
    """
    chunks = []
    with open(md_file) as f:
        line = f.readline()
        while line:
            if not line.startswith(start):
                line = f.readline()
                continue
            chunk = line
            line = f.readline()
            while line and not any(line.startswith(end) for end in ends):
                chunk += line
                line = f.readline()

            chunks.append(chunk)

    return chunks


def parse_legacy_sdks(
    api_md_file: str, model_md_file: str
) -> Tuple[List[LegacySDK], List[LegacyModel]]:
    """
    Parse the legacy SDK markdown files into a list of LegacySDK objects.
    """
    legacy_sdks = [
        LegacySDK.from_md(chunk)
        for chunk in _chunk(api_md_file, "### ", ["### ", "## ", "# "])
    ]
    legacy_models = [
        LegacyModel.from_md(chunk)
        for chunk in _chunk(model_md_file, "## ", ["## ", "# "])
    ]

    return legacy_sdks, legacy_models



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CloudDataBucketPresignedUrlRequest(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'file_type': 'CloudDataBucketFileType',
        'file_name': 'str',
        'access_mode': 'CloudDataBucketAccessMode',
        'scope': 'CloudDataBucketRequestScope'
    }

    attribute_map = {
        'file_type': 'file_type',
        'file_name': 'file_name',
        'access_mode': 'access_mode',
        'scope': 'scope'
    }

    def __init__(self, file_type=None, file_name=None, access_mode=None, scope=None, local_vars_configuration=None):  # noqa: E501
        """CloudDataBucketPresignedUrlRequest - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._file_type = None
        self._file_name = None
        self._access_mode = None
        self._scope = None
        self.discriminator = None

        self.file_type = file_type
        self.file_name = file_name
        self.access_mode = access_mode
        if scope is not None:
            self.scope = scope

    @property
    def file_type(self):
        """Gets the file_type of this CloudDataBucketPresignedUrlRequest.  # noqa: E501

        The type of file that will be uploaded (e.g. 'runtime_env_packages').  # noqa: E501

        :return: The file_type of this CloudDataBucketPresignedUrlRequest.  # noqa: E501
        :rtype: CloudDataBucketFileType
        """
        return self._file_type

    @file_type.setter
    def file_type(self, file_type):
        """Sets the file_type of this CloudDataBucketPresignedUrlRequest.

        The type of file that will be uploaded (e.g. 'runtime_env_packages').  # noqa: E501

        :param file_type: The file_type of this CloudDataBucketPresignedUrlRequest.  # noqa: E501
        :type: CloudDataBucketFileType
        """
        if self.local_vars_configuration.client_side_validation and file_type is None:  # noqa: E501
            raise ValueError("Invalid value for `file_type`, must not be `None`")  # noqa: E501

        self._file_type = file_type

    @property
    def file_name(self):
        """Gets the file_name of this CloudDataBucketPresignedUrlRequest.  # noqa: E501

        The name of the file that will be uploaded (e.g. 'runtime_env_1234.zip'). NOTE: The file name may contain a subdir prefix (eg. my_name/subdir/runtime_env_1234).  # noqa: E501

        :return: The file_name of this CloudDataBucketPresignedUrlRequest.  # noqa: E501
        :rtype: str
        """
        return self._file_name

    @file_name.setter
    def file_name(self, file_name):
        """Sets the file_name of this CloudDataBucketPresignedUrlRequest.

        The name of the file that will be uploaded (e.g. 'runtime_env_1234.zip'). NOTE: The file name may contain a subdir prefix (eg. my_name/subdir/runtime_env_1234).  # noqa: E501

        :param file_name: The file_name of this CloudDataBucketPresignedUrlRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and file_name is None:  # noqa: E501
            raise ValueError("Invalid value for `file_name`, must not be `None`")  # noqa: E501

        self._file_name = file_name

    @property
    def access_mode(self):
        """Gets the access_mode of this CloudDataBucketPresignedUrlRequest.  # noqa: E501

        The access mode for the request (read or write).  # noqa: E501

        :return: The access_mode of this CloudDataBucketPresignedUrlRequest.  # noqa: E501
        :rtype: CloudDataBucketAccessMode
        """
        return self._access_mode

    @access_mode.setter
    def access_mode(self, access_mode):
        """Sets the access_mode of this CloudDataBucketPresignedUrlRequest.

        The access mode for the request (read or write).  # noqa: E501

        :param access_mode: The access_mode of this CloudDataBucketPresignedUrlRequest.  # noqa: E501
        :type: CloudDataBucketAccessMode
        """
        if self.local_vars_configuration.client_side_validation and access_mode is None:  # noqa: E501
            raise ValueError("Invalid value for `access_mode`, must not be `None`")  # noqa: E501

        self._access_mode = access_mode

    @property
    def scope(self):
        """Gets the scope of this CloudDataBucketPresignedUrlRequest.  # noqa: E501

        The scope of the upload request (e.g., organization-level or cloud-level). Defaults to cloud-level.  # noqa: E501

        :return: The scope of this CloudDataBucketPresignedUrlRequest.  # noqa: E501
        :rtype: CloudDataBucketRequestScope
        """
        return self._scope

    @scope.setter
    def scope(self, scope):
        """Sets the scope of this CloudDataBucketPresignedUrlRequest.

        The scope of the upload request (e.g., organization-level or cloud-level). Defaults to cloud-level.  # noqa: E501

        :param scope: The scope of this CloudDataBucketPresignedUrlRequest.  # noqa: E501
        :type: CloudDataBucketRequestScope
        """

        self._scope = scope

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CloudDataBucketPresignedUrlRequest):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CloudDataBucketPresignedUrlRequest):
            return True

        return self.to_dict() != other.to_dict()



from typing import Dict, List, Optional

import tabulate

from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client.models.create_resource_quota import (
    CreateResourceQuota,
)
from anyscale.client.openapi_client.models.list_resource_quotas_query import (
    ListResourceQuotasQuery,
)
from anyscale.client.openapi_client.models.quota import Quota
from anyscale.client.openapi_client.models.resource_quota import ResourceQuota
from anyscale.client.openapi_client.models.resource_quota_status import (
    ResourceQuotaStatus,
)
from anyscale.cloud import get_cloud_id_and_name
from anyscale.controllers.base_controller import BaseController
from anyscale.project import get_project_id_for_cloud_from_name
from anyscale.sdk.anyscale_client.models.page_query import PageQuery
from anyscale.sdk.anyscale_client.models.text_query import TextQuery
from anyscale.utils.user_utils import get_user_id_by_email


class ResourceQuotaController(BaseController):
    def __init__(self):
        super().__init__()
        self.log = LogsLogger()

    def create(  # noqa: PLR0913
        self,
        name: str,
        cloud: str,
        project: Optional[str] = None,
        user_email: Optional[str] = None,
        num_cpus: Optional[int] = None,
        num_instances: Optional[int] = None,
        num_gpus: Optional[int] = None,
        num_accelerators: Optional[Dict[str, int]] = None,
    ) -> str:

        cloud_id, _ = get_cloud_id_and_name(self.api_client, cloud_name=cloud)
        project_id = (
            get_project_id_for_cloud_from_name(
                project, cloud_id, self.api_client, self.anyscale_api_client
            )
            if project
            else None
        )
        user_id = (
            get_user_id_by_email(self.api_client, user_email) if user_email else None
        )

        create_resource_quota = CreateResourceQuota(
            name=name,
            cloud_id=cloud_id,
            project_id=project_id,
            user_id=user_id,
            quota=Quota(
                num_cpus=num_cpus,
                num_instances=num_instances,
                num_gpus=num_gpus,
                num_accelerators=num_accelerators,
            ),
        )

        with self.log.spinner("Creating resource quota..."):
            resource_quota_id = self.api_client.create_resource_quota_api_v2_resource_quotas_post(
                create_resource_quota
            ).result.id

        create_resource_quota_message = [f"Name: {name}\nCloud name: {cloud}"]
        if project:
            create_resource_quota_message.append(f"Project name: {project}")
        if user_email:
            create_resource_quota_message.append(f"User email: {user_email}")
        if num_cpus:
            create_resource_quota_message.append(f"Number of CPUs: {num_cpus}")
        if num_instances:
            create_resource_quota_message.append(
                f"Number of instances: {num_instances}"
            )
        if num_gpus:
            create_resource_quota_message.append(f"Number of GPUs: {num_gpus}")
        if num_accelerators:
            create_resource_quota_message.append(
                f"Number of accelerators: {dict(num_accelerators)}"
            )

        self.log.info("\n".join(create_resource_quota_message))
        self.log.info(f"Resource quota created successfully ID: {resource_quota_id}")

        return resource_quota_id

    def _format_resource_quotas(self, resource_quotas: List[ResourceQuota]) -> str:
        table_rows = []
        for resource_quota in resource_quotas:
            table_rows.append(
                [
                    resource_quota.id,
                    resource_quota.name,
                    resource_quota.cloud_id,
                    resource_quota.project_id,
                    resource_quota.user_id,
                    resource_quota.is_enabled,
                    resource_quota.created_at.strftime("%m/%d/%Y"),
                    resource_quota.deleted_at.strftime("%m/%d/%Y")
                    if resource_quota.deleted_at
                    else None,
                    resource_quota.quota,
                ]
            )
        table = tabulate.tabulate(
            table_rows,
            headers=[
                "ID",
                "NAME",
                "CLOUD ID",
                "PROJECT ID",
                "USER ID",
                "IS ENABLED",
                "CREATED AT",
                "DELETED AT",
                "QUOTA",
            ],
            tablefmt="plain",
        )

        return f"Resource quotas:\n{table}"

    def list_resource_quotas(
        self,
        *,
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        creator_id: Optional[str] = None,
        is_enabled: Optional[bool] = None,
        max_items: int = 20,
    ) -> List[ResourceQuota]:
        cloud_id, _ = (
            get_cloud_id_and_name(self.api_client, cloud_name=cloud)
            if cloud
            else (None, None)
        )

        resource_quotas = self.api_client.search_resource_quotas_api_v2_resource_quotas_search_post(
            ListResourceQuotasQuery(
                name=TextQuery(equals=name) if name else None,
                cloud_id=cloud_id,
                creator_id=creator_id,
                is_enabled=is_enabled,
                paging=PageQuery(count=max_items),
            )
        ).results

        print(self._format_resource_quotas(resource_quotas))
        return resource_quotas

    def delete(self, resource_quota_id: str,) -> None:
        """
        Delete a resource quota.
        """
        with self.log.spinner("Deleting resource quota..."):
            self.api_client.delete_resource_quota_api_v2_resource_quotas_resource_quota_id_delete(
                resource_quota_id
            )

        self.log.info(
            f"Resource quota with ID {resource_quota_id} deleted successfully."
        )

    def set_status(self, resource_quota_id: str, *, is_enabled: bool,) -> None:
        """
        Set the status of a resource quota.
        """
        with self.log.spinner("Setting resource quota status..."):
            self.api_client.set_resource_quota_status_api_v2_resource_quotas_resource_quota_id_status_patch(
                resource_quota_id, ResourceQuotaStatus(is_enabled=is_enabled)
            )

        self.log.info(
            f"{'Enabled' if is_enabled else 'Disabled'} resource quota with ID {resource_quota_id} successfully."
        )



import re

import pytest

from anyscale._private.workload import WorkloadConfig
from anyscale.compute_config.models import ComputeConfig


class TestWorkloadConfig:
    def test_name(self):
        config = WorkloadConfig()
        assert config.name is None

        config = WorkloadConfig(name="my-custom-name")
        assert config.name == "my-custom-name"

        with pytest.raises(TypeError, match="'name' must be a string"):
            WorkloadConfig(name=123)

    def test_image_uri(self):
        config = WorkloadConfig()
        assert config.image_uri is None

        config = WorkloadConfig(image_uri="user/my-custom-image:1")
        assert config.image_uri == "user/my-custom-image:1"

    def test_compute_config(self):
        config = WorkloadConfig()
        assert config.compute_config is None

        config = WorkloadConfig(compute_config="my-custom-compute_config")
        assert config.compute_config == "my-custom-compute_config"

        config = WorkloadConfig(
            compute_config={"cloud": "cloud1", "zones": ["az1", "az2"]}
        )
        assert config.compute_config == ComputeConfig(
            cloud="cloud1", zones=["az1", "az2"]
        )

        config = WorkloadConfig(
            compute_config=ComputeConfig(cloud="cloud2", zones=["az1", "az2"])
        )
        assert config.compute_config == ComputeConfig(
            cloud="cloud2", zones=["az1", "az2"]
        )

        with pytest.raises(
            TypeError,
            match="'compute_config' must be a string, ComputeConfig, or corresponding dict",
        ):
            WorkloadConfig(compute_config=123)

    def test_options(self):
        config = WorkloadConfig()

        options = {
            "name": "test-name",
            "image_uri": "docker.io/libaray/test-image:latest",
            "compute_config": "test-compute-config",
            "excludes": ["some-path"],
        }

        # Test setting fields one at a time.
        for option, val in options.items():
            assert config.options(**{option: val}) == WorkloadConfig(**{option: val})

        # Test setting fields all at once.
        assert config.options(**options) == WorkloadConfig(**options)

    def test_cloud(self):
        config = WorkloadConfig()
        assert config.cloud is None

        config = WorkloadConfig(cloud="my-custom-cloud")
        assert config.cloud == "my-custom-cloud"

        with pytest.raises(TypeError, match="'cloud' must be a string"):
            WorkloadConfig(cloud=123)

    def test_project(self):
        config = WorkloadConfig()
        assert config.project is None

        config = WorkloadConfig(project="my-custom-project")
        assert config.project == "my-custom-project"

        with pytest.raises(TypeError, match="'project' must be a string"):
            WorkloadConfig(project=123)

    def test_invalid_requirements(self):
        WorkloadConfig(requirements="test")
        WorkloadConfig(requirements=["test"])

        with pytest.raises(
            TypeError,
            match=re.escape(
                "'requirements' must be a string (file path) or list of strings."
            ),
        ):
            WorkloadConfig(requirements=1)

        with pytest.raises(
            TypeError,
            match=re.escape(
                "'requirements' must be a string (file path) or list of strings."
            ),
        ):
            WorkloadConfig(requirements=[1])

    def test_invalid_working_dir(self):
        WorkloadConfig(working_dir="test")
        with pytest.raises(TypeError, match="'working_dir' must be a string."):
            WorkloadConfig(working_dir=1)

        with pytest.raises(TypeError, match="'excludes' must be a list of strings."):
            WorkloadConfig(excludes="test")

        with pytest.raises(TypeError, match="'excludes' must be a list of strings."):
            WorkloadConfig(excludes=["test", 1])

    def test_invalid_excludes(self):
        WorkloadConfig(excludes=["test"])
        with pytest.raises(TypeError, match="'excludes' must be a list of strings."):
            WorkloadConfig(excludes="test")

        with pytest.raises(TypeError, match="'excludes' must be a list of strings."):
            WorkloadConfig(excludes=["test", 1])

    def test_invalid_env_vars(self):
        WorkloadConfig(env_vars={"test": "test"})
        with pytest.raises(
            TypeError, match=re.escape("'env_vars' must be a Dict[str, str]."),
        ):
            WorkloadConfig(env_vars={"test": 1})



import pathlib
import subprocess
import tempfile
from typing import Optional, Tuple

import click

from anyscale._private.models.image_uri import ImageURI
from anyscale._private.sdk import _LAZY_SDK_SINGLETONS
from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.util import AnyscaleCommand, convert_kv_strings_to_dict
import anyscale.workspace
from anyscale.workspace._private.workspace_sdk import ANYSCALE_WORKSPACES_SSH_OPTIONS
from anyscale.workspace.commands import _WORKSPACE_SDK_SINGLETON_KEY
from anyscale.workspace.models import (
    UpdateWorkspaceConfig,
    WorkspaceConfig,
    WorkspaceState,
)


log = BlockLogger()  # CLI Logger


def _validate_workspace_name_and_id(
    name: Optional[str], id: Optional[str]  # noqa: A002
):
    if name is None and id is None:
        raise click.ClickException("One of '--name' and '--id' must be provided.")

    if name is not None and id is not None:
        raise click.ClickException("Only one of '--name' and '--id' can be provided.")


@click.group("workspace_v2", help="Anyscale workspace commands V2.")
def workspace_cli() -> None:
    pass


@workspace_cli.command(
    name="create",
    help="Create a workspace on Anyscale.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_CREATE_EXAMPLE,
)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to deploy. When deploying from a file, import path and arguments cannot be provided. Command-line flags will overwrite values read from the file.",
)
@click.option(
    "-n", "--name", required=False, help="Name of the workspace to create.",
)
@click.option(
    "--image-uri",
    required=False,
    default=None,
    type=str,
    help="Container image to use for the workspace. This is exclusive with --containerfile.",
)
@click.option(
    "--registry-login-secret",
    required=False,
    default=None,
    type=str,
    help="Name or identifier of the secret containing credentials to authenticate to the docker registry hosting the image. "
    "This can only be used when 'image_uri' is specified and the image is not hosted on Anyscale.",
)
@click.option(
    "--containerfile",
    required=False,
    default=None,
    type=str,
    help="Path to a containerfile to build the image to use for the workspace. This is exclusive with --image-uri.",
)
@click.option(
    "--ray-version",
    required=False,
    default=None,
    type=str,
    help="The Ray version (X.Y.Z) to the image specified by --image-uri. This is only used when --image-uri is provided. If not provided, the latest Ray version will be used.",
)
@click.option(
    "--compute-config",
    required=False,
    default=None,
    type=str,
    help="Named compute configuration to use for the workspace.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workspace. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the workspace. These will be installed on top of the image.",
)
@click.option(
    "--env",
    required=False,
    multiple=True,
    type=str,
    help="Environment variables to set for the workspace. The format is 'key=value'. This argument can be specified multiple times. When the same key is also specified in the config file, the value from the command-line flag will overwrite the value from the config file.",
)
def create(  # noqa: PLR0913, PLR0912, C901
    config_file: Optional[str],
    name: Optional[str],
    image_uri: Optional[str],
    registry_login_secret: Optional[str],
    ray_version: Optional[str],
    containerfile: Optional[str],
    compute_config: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    requirements: Optional[str],
    env: Optional[Tuple[str]],
) -> None:
    """Creates a new workspace.

    A name must be provided, either in the file or in the arguments.

    `$ anyscale workspace_v2 create -n my-workspace`

    or add all the information in the config file and do:

    `$ anyscale workspace_v2 create -f config-file.yaml`

    Command-line flags override values in the config file.
    """
    if config_file is not None:
        if not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Config file '{config_file}' not found.")

        config = WorkspaceConfig.from_yaml(config_file)
    else:
        config = WorkspaceConfig()

    if containerfile and image_uri:
        raise click.ClickException(
            "Only one of '--containerfile' and '--image-uri' can be provided."
        )

    if ray_version and (not image_uri and not containerfile):
        raise click.ClickException(
            "Ray version can only be used with an image or containerfile.",
        )

    if registry_login_secret and (
        not image_uri or ImageURI.from_str(image_uri).is_cluster_env_image()
    ):
        raise click.ClickException(
            "Registry login secret can only be used with an image that is not hosted on Anyscale."
        )

    if name is not None:
        config = config.options(name=name)

    if not config.name:
        raise click.ClickException("Workspace name must be configured")

    if image_uri is not None:
        config = config.options(image_uri=image_uri)

    if registry_login_secret is not None:
        config = config.options(registry_login_secret=registry_login_secret)

    if ray_version is not None:
        config = config.options(ray_version=ray_version)

    if containerfile is not None:
        config = config.options(containerfile=containerfile)

    if compute_config is not None:
        config = config.options(compute_config=compute_config)

    if cloud is not None:
        config = config.options(cloud=cloud)
    if project is not None:
        config = config.options(project=project)

    if requirements is not None:
        if not pathlib.Path(requirements).is_file():
            raise click.ClickException(f"Requirements file '{requirements}' not found.")
        config = config.options(requirements=requirements)
    if env:
        env_dict = convert_kv_strings_to_dict(env)
        if env_dict:
            config = config.options(env_vars=env_dict)

    anyscale.workspace.create(config,)


@workspace_cli.command(
    name="start",
    short_help="Starts a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_START_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
def start(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Start a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.start(name=name, id=id, cloud=cloud, project=project)


@workspace_cli.command(
    name="terminate",
    short_help="Terminate a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_TERMINATE_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
def terminate(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Terminate a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.terminate(name=name, id=id, cloud=cloud, project=project)


@workspace_cli.command(
    name="status",
    short_help="Get the status of a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_STATUS_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
def status(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Get the status of a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    status = anyscale.workspace.status(name=name, id=id, cloud=cloud, project=project)
    log.info(status)


@workspace_cli.command(
    name="wait",
    short_help="Wait for a workspace to reach a certain status.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_WAIT_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.option(
    "--timeout-s",
    required=False,
    default=1800,
    type=float,
    help="The maximum time in seconds to wait for the workspace to reach the desired state. Default to 30 minutes.",
)
@click.option(
    "--state",
    required=False,
    default=WorkspaceState.RUNNING,
    type=str,
    help="The desired terminal state to wait for. Default is 'RUNNING'.",
)
def wait(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    timeout_s: float,
    state: str,
) -> None:
    """Wait for a workspace to reach a terminal state.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    try:
        state = WorkspaceState.validate(state)
    except ValueError as e:
        raise click.ClickException(str(e))
    anyscale.workspace.wait(
        name=name, id=id, cloud=cloud, project=project, timeout_s=timeout_s, state=state
    )


@workspace_cli.command(
    name="ssh",
    short_help="SSH into a workspace.",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_SSH_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.pass_context
def ssh(
    ctx,
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """SSH into a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    You may pass extra args for the ssh command, for example to setup port forwarding:
    anyscale workspace_v2 ssh -n workspace-name -- -L 9000:localhost:9000
    """

    _validate_workspace_name_and_id(name=name, id=id)
    assert (
        anyscale.workspace.status(name=name, id=id, cloud=cloud, project=project)
        == WorkspaceState.RUNNING
    ), "Workspace must be running to SSH into it."
    workspace_private_sdk = _LAZY_SDK_SINGLETONS[_WORKSPACE_SDK_SINGLETON_KEY]
    dir_name = workspace_private_sdk.get_default_dir_name(
        name=name, id=id, cloud=cloud, project=project
    )
    command = f"cd {dir_name} && /bin/bash"
    with tempfile.TemporaryDirectory() as tmpdirname:
        host_name, config_file = anyscale.workspace.generate_ssh_config_file(
            name=name, id=id, cloud=cloud, project=project, ssh_config_path=tmpdirname
        )
        args = ctx.args
        ssh_command = (
            ["ssh"]
            + ANYSCALE_WORKSPACES_SSH_OPTIONS
            + [host_name]
            + ["-F", config_file]
            + ["-tt"]
            + (args if args and len(args) > 0 else [""])
            + [f"bash -i -c {command}"]
        )

        subprocess.run(ssh_command, check=False)


@workspace_cli.command(
    name="run_command",
    short_help="Run a command in a workspace.",
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_RUN_COMMAND_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.argument("command", type=str)
def run_command(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    command: str,
) -> None:
    """Run a command in a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.run_command(
        name=name, id=id, cloud=cloud, project=project, command=command
    )


@workspace_cli.command(
    name="pull",
    short_help="Pull the working directory of a workspace.",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_PULL_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.option(
    "--local-dir",
    required=False,
    default=None,
    type=str,
    help="Local directory to pull the workspace directory to. If not provided, the current directory will be used.",
)
@click.option(
    "--pull-git-state",
    required=False,
    default=False,
    is_flag=True,
    help="Pull the git state of the workspace.",
)
@click.pass_context
def pull(
    ctx,
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    local_dir: Optional[str],
    pull_git_state: bool = False,
) -> None:
    """Pull the working directory of a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    This command depends on rsync, please make sure it is installed on your system.

    You may pass extra args for the rsync command, for example to exclude files:
    anyscale workspace_v2 pull -n workspace-name -- --exclude='log.txt'
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.pull(
        name=name,
        id=id,
        cloud=cloud,
        project=project,
        local_dir=local_dir,
        pull_git_state=pull_git_state,
        rsync_args=ctx.args,
    )


@workspace_cli.command(
    name="push",
    short_help="Push a local directory to a workspace.",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
    cls=AnyscaleCommand,
    example=command_examples.WORKSPACE_PUSH_EXAMPLE,
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.option(
    "--local-dir",
    required=False,
    default=None,
    type=str,
    help="Local directory to push to the workspace. If not provided, the current directory will be used.",
)
@click.option(
    "--push-git-state",
    required=False,
    default=False,
    is_flag=True,
    help="Push the git state of the workspace.",
)
@click.pass_context
def push(
    ctx,
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    local_dir: Optional[str],
    push_git_state: bool = False,
) -> None:
    """Push a local directory to a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    This command depends on rsync, please make sure it is installed on your system.

    You may pass extra args for the rsync command, for example to exclude files:
    anyscale workspace_v2 push -n workspace-name -- --exclude='log.txt'
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.push(
        name=name,
        id=id,
        cloud=cloud,
        project=project,
        local_dir=local_dir,
        push_git_state=push_git_state,
        rsync_args=ctx.args,
    )


@workspace_cli.command(
    name="update", help="Update an existing workspace on Anyscale.",
)
@click.argument("workspace-id", type=str)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to update. Command-line flags will overwrite values read from the file. Unspecified fields will retain their current values, while specified fields will be updated.",
)
@click.option(
    "-n", "--name", required=False, help="New name of the workspace.",
)
@click.option(
    "--image-uri",
    required=False,
    default=None,
    type=str,
    help="New container image to use for the workspace. This is exclusive with --containerfile.",
)
@click.option(
    "--registry-login-secret",
    required=False,
    default=None,
    type=str,
    help="Name or identifier of the secret containing credentials to authenticate to the docker registry hosting the image. "
    "This can only be used when 'image_uri' is specified and the image is not hosted on Anyscale.",
)
@click.option(
    "--containerfile",
    required=False,
    default=None,
    type=str,
    help="Path to a containerfile to build the image to use for the workspace. This is exclusive with --image-uri.",
)
@click.option(
    "--ray-version",
    required=False,
    default=None,
    type=str,
    help="New Ray version (X.Y.Z) to use with the image specified by --image-uri. This is only used when --image-uri is provided. If not provided, the latest Ray version will be used.",
)
@click.option(
    "--compute-config",
    required=False,
    default=None,
    type=str,
    help="New named compute configuration to use for the workspace.",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the workspace. These will be installed on top of the image.",
)
@click.option(
    "--env",
    required=False,
    multiple=True,
    type=str,
    help="New environment variables to set for the workspace. The format is 'key=value'. This argument can be specified multiple times. When the same key is also specified in the config file, the value from the command-line flag will overwrite the value from the config file.",
)
def update(  # noqa: PLR0913, PLR0912
    workspace_id: str,
    config_file: Optional[str],
    name: Optional[str],
    image_uri: Optional[str],
    registry_login_secret: Optional[str],
    ray_version: Optional[str],
    containerfile: Optional[str],
    compute_config: Optional[str],
    requirements: Optional[str],
    env: Optional[Tuple[str]],
) -> None:
    """Updates an existing workspace.

    Example:
    `$ anyscale workspace_v2 update <workspace-id> --name new-name`

    Command-line flags override values in the config file.
    Unspecified fields will retain their current values, while specified fields will be updated.
    """

    if config_file is not None:
        if not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Config file '{config_file}' not found.")

        config = UpdateWorkspaceConfig.from_yaml(config_file)
    else:
        config = UpdateWorkspaceConfig()

    if containerfile and image_uri:
        raise click.ClickException(
            "Only one of '--containerfile' and '--image-uri' can be provided."
        )

    if ray_version and (not image_uri and not containerfile):
        raise click.ClickException(
            "Ray version can only be used with an image or containerfile.",
        )

    if registry_login_secret and (
        not image_uri or ImageURI.from_str(image_uri).is_cluster_env_image()
    ):
        raise click.ClickException(
            "Registry login secret can only be used with an image that is not hosted on Anyscale."
        )

    if name is not None:
        config = config.options(name=name)

    if image_uri is not None:
        config = config.options(image_uri=image_uri)

    if registry_login_secret is not None:
        config = config.options(registry_login_secret=registry_login_secret)

    if ray_version is not None:
        config = config.options(ray_version=ray_version)

    if containerfile is not None:
        config = config.options(containerfile=containerfile)

    if compute_config is not None:
        config = config.options(compute_config=compute_config)

    if requirements is not None:
        if not pathlib.Path(requirements).is_file():
            raise click.ClickException(f"Requirements file '{requirements}' not found.")
        config = config.options(requirements=requirements)

    if env:
        env_dict = convert_kv_strings_to_dict(env)
        if env_dict:
            config = config.options(env_vars=env_dict)

    # Apply the update
    anyscale.workspace.update(id=workspace_id, config=config)



from typing import Tuple

from common import TestLogger
import pytest

from anyscale import Anyscale
from anyscale._private.anyscale_client.fake_anyscale_client import FakeAnyscaleClient
from anyscale._private.models.image_uri import ImageURI
from anyscale._private.sdk.timer import FakeTimer
from anyscale.image import ImageSDK
from anyscale.image.models import ImageBuild, ImageBuildStatus


@pytest.fixture()
def image_sdk_with_fakes(
    sdk_with_fakes: Tuple[Anyscale, FakeAnyscaleClient, TestLogger, FakeTimer]
) -> Tuple[ImageSDK, FakeAnyscaleClient, TestLogger, FakeTimer]:
    sdk, client, logger, timer = sdk_with_fakes
    return sdk.image, client, logger, timer


class TestBuildImage:
    def test_build_image(
        self,
        image_sdk_with_fakes: Tuple[
            ImageSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = image_sdk_with_fakes
        container_file = "\n".join(
            [
                "FROM docker.io/my/sofian-base-image:latest",
                'RUN pip install "requests"',
                'RUN pip install "flask"',
            ]
        )

        # call the SDK
        image_uri = sdk.build(container_file, name="bldname123", ray_version="2.24.0")

        assert image_uri, "Image URI is None"
        cluster_env_build_id = fake_client.get_cluster_env_build_id_from_image_uri(
            ImageURI.from_str(image_uri), ray_version="2.24.0"
        )
        assert fake_client._builds[cluster_env_build_id].containerfile == container_file


class TestGetImage:
    def test_get_image(
        self,
        image_sdk_with_fakes: Tuple[
            ImageSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = image_sdk_with_fakes
        fake_client.get_cluster_env_build_id_from_containerfile(
            "my-image", "FROM python:3.8", anonymous=False, ray_version="2.24.0"
        )

        # call the sdk
        img_build = sdk.get(name="my-image")

        assert img_build == ImageBuild(
            status=ImageBuildStatus.SUCCEEDED,
            uri="anyscale/image/my-image:1",
            ray_version="2.24.0",
        )



from typing import Optional

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk import sdk_docs
from anyscale._private.sdk.base_sdk import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.schedule._private.schedule_sdk import PrivateScheduleSDK
from anyscale.schedule.commands import (
    _APPLY_ARG_DOCSTRINGS,
    _APPLY_EXAMPLE,
    _SET_STATE_ARG_DOCSTRINGS,
    _SET_STATE_EXAMPLE,
    _STATUS_ARG_DOCSTRINGS,
    _STATUS_EXAMPLE,
    _TRIGGER_ARG_DOCSTRINGS,
    _TRIGGER_EXAMPLE,
    apply,
    set_state,
    status,
    trigger,
)
from anyscale.schedule.models import ScheduleConfig, ScheduleState, ScheduleStatus


class ScheduleSDK:
    def __init__(
        self,
        *,
        client: Optional[AnyscaleClientInterface] = None,
        logger: Optional[BlockLogger] = None,
        timer: Optional[Timer] = None,
    ):
        self._private_sdk = PrivateScheduleSDK(
            client=client, logger=logger, timer=timer
        )

    @sdk_docs(
        doc_py_example=_APPLY_EXAMPLE, arg_docstrings=_APPLY_ARG_DOCSTRINGS,
    )
    def apply(self, config: ScheduleConfig,) -> str:  # noqa: F811
        """Apply or update a schedule.

        Returns the id of the schedule.
        """
        return self._private_sdk.apply(config=config)

    @sdk_docs(
        doc_py_example=_SET_STATE_EXAMPLE, arg_docstrings=_SET_STATE_ARG_DOCSTRINGS,
    )
    def set_state(  # noqa: F811
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        state: ScheduleState,
    ) -> str:  # noqa: F811
        """Set the state of a schedule.

        Returns the id of the schedule.
        """
        return self._private_sdk.set_state(
            id=id, name=name, cloud=cloud, project=project, state=state,
        )

    @sdk_docs(doc_py_example=_STATUS_EXAMPLE, arg_docstrings=_STATUS_ARG_DOCSTRINGS)
    def status(  # noqa: F811
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> ScheduleStatus:
        """Return the status of the schedule.
        """
        return self._private_sdk.status(id=id, name=name, cloud=cloud, project=project)

    @sdk_docs(doc_py_example=_TRIGGER_EXAMPLE, arg_docstrings=_TRIGGER_ARG_DOCSTRINGS)
    def trigger(  # noqa: F811
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        """Trigger the execution of the schedule.
        """
        return self._private_sdk.trigger(id=id, name=name, cloud=cloud, project=project)



from functools import partial
import importlib
import re
import subprocess
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from click import ClickException
from google.api_core.exceptions import Forbidden, NotFound, PermissionDenied
import google.auth
from google.auth.credentials import Credentials, CredentialsWithQuotaProject
from google.auth.exceptions import DefaultCredentialsError, RefreshError
from google.auth.transport.requests import Request
import google.cloud
from google.iam.v1.policy_pb2 import Binding
from google.oauth2 import service_account
from googleapiclient.discovery import build as api_client_build
from googleapiclient.errors import HttpError

from anyscale.cli_logger import CloudSetupLogger
from anyscale.client.openapi_client.models.cloud_analytics_event_cloud_resource import (
    CloudAnalyticsEventCloudResource,
)
from anyscale.client.openapi_client.models.gcp_file_store_config import (
    GCPFileStoreConfig,
)
from anyscale.client.openapi_client.models.gcp_memorystore_instance_config import (
    GCPMemorystoreInstanceConfig,
)
from anyscale.util import confirm
from anyscale.utils.cloud_utils import CloudSetupError


GCP_REQUIRED_APIS = [
    "compute.googleapis.com",  # Compute Engine
    "file.googleapis.com",  # Filestore
    "storage-component.googleapis.com",  # Cloud Storage
    "storage.googleapis.com",  # Cloud Storage
    "deploymentmanager.googleapis.com",  # Deployment Manager
    "cloudresourcemanager.googleapis.com",  # Resource Manager
    "certificatemanager.googleapis.com",  # Certificate Manager for Services
]

GCP_RESOURCE_DICT: Dict[str, CloudAnalyticsEventCloudResource] = {
    "VPC": CloudAnalyticsEventCloudResource.GCP_VPC,
    "Subnet": CloudAnalyticsEventCloudResource.GCP_SUBNET,
    "Project": CloudAnalyticsEventCloudResource.GCP_PROJECT,
    "Anyscale Access Service Account": CloudAnalyticsEventCloudResource.GCP_SERVICE_ACCOUNT,
    "Dataplane Service Account": CloudAnalyticsEventCloudResource.GCP_SERVICE_ACCOUNT,
    "Firewall Policy": CloudAnalyticsEventCloudResource.GCP_FIREWALL_POLICY,
    "Filestore": CloudAnalyticsEventCloudResource.GCP_FILESTORE,
    "Google Cloud Storage Bucket": CloudAnalyticsEventCloudResource.GCP_STORAGE_BUCKET,
    "Memorystore": CloudAnalyticsEventCloudResource.GCP_MEMORYSTORE,
}


class GCPLogger:
    def __init__(
        self, logger: CloudSetupLogger, project_id: str, spinner: Any, yes: bool = False
    ):
        self.internal = logger
        self.project_id = project_id
        self.spinner = spinner
        self.yes = yes

    def log_resource_not_found_error(
        self, resource_name: str, resource_id: str, project_id: Optional[str] = None,
    ):
        if resource_name == "Project":
            self.internal.error(
                f"Could not find {resource_name} with id {resource_id}. Please validate that you're using the correct GCP project."
            )
        else:
            project_id = project_id or self.project_id
            self.internal.error(
                f"Could not find {resource_name} with id {resource_id} in project {project_id}. Please validate that you're using the correct GCP project and that the resource values are correct."
            )
        resource = GCP_RESOURCE_DICT.get(resource_name)
        if resource:
            self.internal.log_resource_error(
                resource, CloudSetupError.RESOURCE_NOT_FOUND
            )

    def confirm_missing_permission(self, error_str: str):
        self.internal.error(error_str)
        self.spinner.stop()
        confirm(
            "If the service account has these required permissions granted via other roles\n"
            "or via a group, press 'y' to continue with verification or 'N' to abort.",
            yes=self.yes,
        )
        self.spinner.start()


class GoogleCloudClientFactory:
    """Factory to generate both Google Cloud Client libraries & Google API Client libraries.

    Google Cloud Client libraries are instantiated by:
    ```
        factory = GoogleCloudClientFactory(credentials=AnonymousCredentials())
        client = factory.compute_v1.ExampleClient()
    ```

    Google API Client libraries are instantiated by:
    ```
        factory = GoogleCloudClientFactory(credentials=AnonymousCredentials())
        client = factory.build("iam", "v1")
    ```
    """

    def __init__(self, credentials: Credentials, force_rest=False, **kwargs):
        kwargs["credentials"] = credentials
        self.kwargs = kwargs
        self.force_rest = force_rest

    def __getattr__(self, client_library: str):
        """Get a wrapped Google Cloud Client library that injects default values from the factory."""
        if not hasattr(google.cloud, client_library):
            importlib.import_module(f"google.cloud.{client_library}")
        module = getattr(google.cloud, client_library)
        kwargs = self.kwargs.copy()

        # NOTE the `storage` library only supports HTTP, but doesn't
        # have the `transport` argument in its signature.
        if self.force_rest and client_library != "storage":
            kwargs["transport"] = "rest"

        class WrappedClient:
            def __getattr__(self, client_type: str):
                return partial(getattr(module, client_type), **kwargs)

        return WrappedClient()

    def build(self, service_name: str, version: str):
        """Return a Google API Client with default values from the factor"""
        return api_client_build(
            service_name, version, cache_discovery=False, **self.kwargs
        )


def get_application_default_credentials(
    logger: CloudSetupLogger,
) -> Tuple[Credentials, Optional[str]]:
    """Get application default credentials, or run `gcloud` to try to log in."""
    try:
        credentials, project = google.auth.default(
            default_scopes="https://www.googleapis.com/auth/cloud-platform"
        )
        # Don't try to refresh credentials if they are from a Service Account Key File
        if isinstance(credentials, service_account.Credentials):
            return credentials, project
        try:
            credentials.refresh(Request())
            return credentials, project
        except RefreshError:
            logger.warning(
                "Reauthentication is needed, running `gcloud auth application-default login`"
            )
    except DefaultCredentialsError:
        logger.warning(
            "Could not automatically determine Google Application Default Credentials, trying to authenticate via GCloud"
        )
    auth_login = subprocess.run(
        ["gcloud", "auth", "application-default", "login"], check=False
    )
    if auth_login.returncode != 0:
        raise RuntimeError("Failed to authenticate via gcloud")

    return google.auth.default()


def get_google_cloud_client_factory(logger: CloudSetupLogger, project_id: str):
    credentials, credentials_project = get_application_default_credentials(logger)

    # Add a trivial call to the API to verify that the credentials are valid
    try:
        factory = GoogleCloudClientFactory(credentials=credentials)
        factory.build("serviceusage", "v1").services().get(
            name=f"projects/{project_id}/services/serviceusage.googleapis.com"
        ).execute()
    except HttpError as e:
        logger.error(e)
        raise ClickException(
            f"Failed to get project {project_id}. \n"
            "Please make sure:\n"
            "1) the project exists\n"
            "2) the account you're using has the `owner` role on the project\n"
            "3) the credentials are valid (run `gcloud auth application-default login` to reauthenticate)"
        )

    # swtich quota project to given project
    if credentials_project != project_id:
        logger.info(
            f"Default credentials are for {credentials_project}, but this cloud is being configured for {project_id}.\n"
            f"Switching quota project to {project_id}"
        )
        if isinstance(credentials, CredentialsWithQuotaProject):
            credentials = credentials.with_quota_project(project_id)
            factory = GoogleCloudClientFactory(credentials=credentials)

    return factory


def binding_from_dictionary(
    inp: List[Dict[str, Union[List[str], str]]]
) -> List[Binding]:
    return [Binding(role=b["role"], members=b["members"]) for b in inp]


def check_policy_bindings(
    iam_policy: List[Binding], member: str, possible_roles: Set[str]
) -> bool:
    """Checks if `member` has any role in `possible_roles` given the specified iam_policy."""
    return any(
        policy.role in possible_roles and member in policy.members
        for policy in iam_policy
    )


def check_required_policy_bindings(
    iam_policy: List[Binding], member: str, required_roles: Set[str]
) -> bool:
    """Checks if `member` has all roles in `required_roles` given the specified iam_policy."""
    granted_roles = {policy.role for policy in iam_policy if member in policy.members}
    return required_roles.issubset(granted_roles)


def get_gcp_filestore_config(
    factory: GoogleCloudClientFactory,
    project_id: str,
    vpc_name: str,
    filestore_location: str,
    filestore_instance_id: str,
    logger: CloudSetupLogger,
):
    client = factory.filestore_v1.CloudFilestoreManagerClient()
    instance_name = "projects/{}/locations/{}/instances/{}".format(
        project_id, filestore_location, filestore_instance_id
    )
    try:
        file_store = client.get_instance(name=instance_name)
    except NotFound as e:
        logger.log_resource_error(
            CloudAnalyticsEventCloudResource.GCP_FILESTORE,
            CloudSetupError.RESOURCE_NOT_FOUND,
        )
        raise ClickException(
            f"Could not find Filestore with id {filestore_instance_id} in project {project_id}. Please validate that you're using the correct GCP project and that the resource values are correct. Error details: {e}"
        )
    root_dir = file_store.file_shares[0].name
    for v in file_store.networks:
        # v.network can be <vpc_name> or projects/<project_number>/global/networks/<vpc_name>
        network_name = v.network.split("/")[-1]
        if vpc_name == network_name:
            mount_target_ip = v.ip_addresses[0]
            break
    else:
        logger.error(
            f"Filestore {filestore_instance_id} is not connected to {vpc_name}, but to {[v.network for v in file_store.networks]}. "
            f"This cannot be edited on an existing Filestore instance. Please recreate the filestore and connect it to {vpc_name}."
        )
        logger.log_resource_error(
            CloudAnalyticsEventCloudResource.GCP_FILESTORE,
            CloudSetupError.FILESTORE_NOT_CONNECTED_TO_VPC,
        )
        raise ClickException(
            f"Filestore {filestore_instance_id} is not connected to {vpc_name}."
        )
    return GCPFileStoreConfig(
        instance_name=instance_name, root_dir=root_dir, mount_target_ip=mount_target_ip,
    )


def get_filestore_location_and_instance_id(
    gcp_filestore_config: GCPFileStoreConfig,
) -> Tuple[str, str]:
    instance_name = gcp_filestore_config.instance_name
    # instance name follows format of "projects/{}/locations/{}/instances/{}"
    pattern = (
        r"projects/[^/]+/locations/(?P<location>[^/]+)/instances/(?P<instance_id>[^/]+)"
    )
    match = re.match(pattern, instance_name)

    if match:
        filestore_location = match.group("location")
        filestore_instance_id = match.group("instance_id")
        return filestore_location, filestore_instance_id
    else:
        raise ClickException(
            f"Could not parse Filestore instance name {instance_name}."
        )


def get_gcp_memorystore_config(
    factory: GoogleCloudClientFactory, instance_name: Optional[str]
) -> Optional[GCPMemorystoreInstanceConfig]:
    """Get the Memorystore instance config from GCP
    Returns None if instance_name is None
    """
    if not instance_name:
        return None

    client = factory.redis_v1.CloudRedisClient()

    try:
        instance = client.get_instance(name=instance_name)
    except (NotFound, PermissionDenied, Forbidden) as e:
        raise ClickException(
            f"Error occurred when trying to access the memorystore instance {instance_name}: {e}.\nPlease validate that you're using the correct GCP project and that the resource values are correct."
        )
    return GCPMemorystoreInstanceConfig(
        name=instance_name, endpoint=instance.host + ":" + str(instance.port),
    )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class JobQueue(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'name': 'str',
        'user_provided_id': 'str',
        'created_at': 'datetime',
        'creator_id': 'str',
        'project_id': 'str',
        'cloud_id': 'str',
        'compute_config_id': 'str',
        'cluster_environment_build_id': 'str',
        'max_concurrency': 'int',
        'idle_timeout_sec': 'int',
        'execution_mode': 'JobQueueExecutionMode',
        'state': 'JobQueueState'
    }

    attribute_map = {
        'id': 'id',
        'name': 'name',
        'user_provided_id': 'user_provided_id',
        'created_at': 'created_at',
        'creator_id': 'creator_id',
        'project_id': 'project_id',
        'cloud_id': 'cloud_id',
        'compute_config_id': 'compute_config_id',
        'cluster_environment_build_id': 'cluster_environment_build_id',
        'max_concurrency': 'max_concurrency',
        'idle_timeout_sec': 'idle_timeout_sec',
        'execution_mode': 'execution_mode',
        'state': 'state'
    }

    def __init__(self, id=None, name=None, user_provided_id=None, created_at=None, creator_id=None, project_id=None, cloud_id=None, compute_config_id=None, cluster_environment_build_id=None, max_concurrency=1, idle_timeout_sec=None, execution_mode=None, state=None, local_vars_configuration=None):  # noqa: E501
        """JobQueue - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._name = None
        self._user_provided_id = None
        self._created_at = None
        self._creator_id = None
        self._project_id = None
        self._cloud_id = None
        self._compute_config_id = None
        self._cluster_environment_build_id = None
        self._max_concurrency = None
        self._idle_timeout_sec = None
        self._execution_mode = None
        self._state = None
        self.discriminator = None

        self.id = id
        if name is not None:
            self.name = name
        if user_provided_id is not None:
            self.user_provided_id = user_provided_id
        self.created_at = created_at
        self.creator_id = creator_id
        self.project_id = project_id
        self.cloud_id = cloud_id
        self.compute_config_id = compute_config_id
        self.cluster_environment_build_id = cluster_environment_build_id
        if max_concurrency is not None:
            self.max_concurrency = max_concurrency
        self.idle_timeout_sec = idle_timeout_sec
        if execution_mode is not None:
            self.execution_mode = execution_mode
        self.state = state

    @property
    def id(self):
        """Gets the id of this JobQueue.  # noqa: E501

        The id of this job queue  # noqa: E501

        :return: The id of this JobQueue.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this JobQueue.

        The id of this job queue  # noqa: E501

        :param id: The id of this JobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this JobQueue.  # noqa: E501

        Name of the job queue  # noqa: E501

        :return: The name of this JobQueue.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this JobQueue.

        Name of the job queue  # noqa: E501

        :param name: The name of this JobQueue.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def user_provided_id(self):
        """Gets the user_provided_id of this JobQueue.  # noqa: E501

        Optional user-provided identifier of the queue that could be subsequently used to reference the queue when submitting jobs.   # noqa: E501

        :return: The user_provided_id of this JobQueue.  # noqa: E501
        :rtype: str
        """
        return self._user_provided_id

    @user_provided_id.setter
    def user_provided_id(self, user_provided_id):
        """Sets the user_provided_id of this JobQueue.

        Optional user-provided identifier of the queue that could be subsequently used to reference the queue when submitting jobs.   # noqa: E501

        :param user_provided_id: The user_provided_id of this JobQueue.  # noqa: E501
        :type: str
        """

        self._user_provided_id = user_provided_id

    @property
    def created_at(self):
        """Gets the created_at of this JobQueue.  # noqa: E501

        The time this job queue was created  # noqa: E501

        :return: The created_at of this JobQueue.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this JobQueue.

        The time this job queue was created  # noqa: E501

        :param created_at: The created_at of this JobQueue.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def creator_id(self):
        """Gets the creator_id of this JobQueue.  # noqa: E501

        The id of the user who created this job queue  # noqa: E501

        :return: The creator_id of this JobQueue.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this JobQueue.

        The id of the user who created this job queue  # noqa: E501

        :param creator_id: The creator_id of this JobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def project_id(self):
        """Gets the project_id of this JobQueue.  # noqa: E501

        The project id to which the job queue belongs to.  # noqa: E501

        :return: The project_id of this JobQueue.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this JobQueue.

        The project id to which the job queue belongs to.  # noqa: E501

        :param project_id: The project_id of this JobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def cloud_id(self):
        """Gets the cloud_id of this JobQueue.  # noqa: E501

        The cloud id to which the job queue belongs to.  # noqa: E501

        :return: The cloud_id of this JobQueue.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this JobQueue.

        The cloud id to which the job queue belongs to.  # noqa: E501

        :param cloud_id: The cloud_id of this JobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def compute_config_id(self):
        """Gets the compute_config_id of this JobQueue.  # noqa: E501

        The id of the compute configuration that will be used to create cluster associated with the queue.  # noqa: E501

        :return: The compute_config_id of this JobQueue.  # noqa: E501
        :rtype: str
        """
        return self._compute_config_id

    @compute_config_id.setter
    def compute_config_id(self, compute_config_id):
        """Sets the compute_config_id of this JobQueue.

        The id of the compute configuration that will be used to create cluster associated with the queue.  # noqa: E501

        :param compute_config_id: The compute_config_id of this JobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and compute_config_id is None:  # noqa: E501
            raise ValueError("Invalid value for `compute_config_id`, must not be `None`")  # noqa: E501

        self._compute_config_id = compute_config_id

    @property
    def cluster_environment_build_id(self):
        """Gets the cluster_environment_build_id of this JobQueue.  # noqa: E501

        The id of the cluster environment build that will be used to create cluster associated with the queue.  # noqa: E501

        :return: The cluster_environment_build_id of this JobQueue.  # noqa: E501
        :rtype: str
        """
        return self._cluster_environment_build_id

    @cluster_environment_build_id.setter
    def cluster_environment_build_id(self, cluster_environment_build_id):
        """Sets the cluster_environment_build_id of this JobQueue.

        The id of the cluster environment build that will be used to create cluster associated with the queue.  # noqa: E501

        :param cluster_environment_build_id: The cluster_environment_build_id of this JobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cluster_environment_build_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_environment_build_id`, must not be `None`")  # noqa: E501

        self._cluster_environment_build_id = cluster_environment_build_id

    @property
    def max_concurrency(self):
        """Gets the max_concurrency of this JobQueue.  # noqa: E501

        Max number of jobs to be run concurrently. Defaults to 1, ie running no more than 1 job at a time.  # noqa: E501

        :return: The max_concurrency of this JobQueue.  # noqa: E501
        :rtype: int
        """
        return self._max_concurrency

    @max_concurrency.setter
    def max_concurrency(self, max_concurrency):
        """Sets the max_concurrency of this JobQueue.

        Max number of jobs to be run concurrently. Defaults to 1, ie running no more than 1 job at a time.  # noqa: E501

        :param max_concurrency: The max_concurrency of this JobQueue.  # noqa: E501
        :type: int
        """

        self._max_concurrency = max_concurrency

    @property
    def idle_timeout_sec(self):
        """Gets the idle_timeout_sec of this JobQueue.  # noqa: E501

        Max period of time queue will be accepting new jobs, before being sealed off and its associated cluster being shutdown  # noqa: E501

        :return: The idle_timeout_sec of this JobQueue.  # noqa: E501
        :rtype: int
        """
        return self._idle_timeout_sec

    @idle_timeout_sec.setter
    def idle_timeout_sec(self, idle_timeout_sec):
        """Sets the idle_timeout_sec of this JobQueue.

        Max period of time queue will be accepting new jobs, before being sealed off and its associated cluster being shutdown  # noqa: E501

        :param idle_timeout_sec: The idle_timeout_sec of this JobQueue.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and idle_timeout_sec is None:  # noqa: E501
            raise ValueError("Invalid value for `idle_timeout_sec`, must not be `None`")  # noqa: E501

        self._idle_timeout_sec = idle_timeout_sec

    @property
    def execution_mode(self):
        """Gets the execution_mode of this JobQueue.  # noqa: E501

        Execution mode of the jobs submitted into the queue (one of: FIFO,LIFO,PRIORITY  # noqa: E501

        :return: The execution_mode of this JobQueue.  # noqa: E501
        :rtype: JobQueueExecutionMode
        """
        return self._execution_mode

    @execution_mode.setter
    def execution_mode(self, execution_mode):
        """Sets the execution_mode of this JobQueue.

        Execution mode of the jobs submitted into the queue (one of: FIFO,LIFO,PRIORITY  # noqa: E501

        :param execution_mode: The execution_mode of this JobQueue.  # noqa: E501
        :type: JobQueueExecutionMode
        """

        self._execution_mode = execution_mode

    @property
    def state(self):
        """Gets the state of this JobQueue.  # noqa: E501

        The current state of this job queue  # noqa: E501

        :return: The state of this JobQueue.  # noqa: E501
        :rtype: JobQueueState
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this JobQueue.

        The current state of this job queue  # noqa: E501

        :param state: The state of this JobQueue.  # noqa: E501
        :type: JobQueueState
        """
        if self.local_vars_configuration.client_side_validation and state is None:  # noqa: E501
            raise ValueError("Invalid value for `state`, must not be `None`")  # noqa: E501

        self._state = state

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, JobQueue):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, JobQueue):
            return True

        return self.to_dict() != other.to_dict()
